{**
*
*         Script Reporter Component library
*         B-tree builder for the expressions
*          Vlad Ilushchenko 1998-99
*
*
**}

unit ExprParser;

interface

uses Tokenizer, SysUtils, Reporter;

type
  TRepExprNode = class;

  TRepExprNode = class
  public
    ExprName: String;
    Left: TRepExprNode;
    Right: TRepExprNode;
    Line: Integer;
    Col:  Integer;
    constructor Create(AName: String);
  end;

  TRepExprParser = class
  private
    FStreamParser: TStreamTokenizer;
    FReporter: TCustomScriptRep;
    FRootNode: TRepExprNode;
    FBreakSymbols: String;
    function ParseNextExpr(ANode: TRepExprNode;
      TokenPriority: Integer): TRepExprNode;
    function GetTokenPriority(Token: TToken): Integer;
    function TokenBreak(ATokenText: String): Boolean;
  public
    constructor Create(AStreamParser: TStreamTokenizer; Reporter: TCustomScriptRep);
    destructor Destroy; override;

    procedure BuildTree(BreakSymbols: String);
    property RootNode: TRepExprNode read FRootNode;
  end;

procedure UtilizeTree(RootNode: TRepExprNode);

implementation

const
  ExprNameArray: array [0..19] of string =
       ('->', '(', 'NOT', '!', '^', '*', '/', '+', '-', '<', '>',
        '<=', '>=', '=', '<>', 'AND', 'OR',  ',', ')', ';');

  ExprPriorityArray: array [0..19] of integer =
       ( 21,   20,   19,  13,  12,  10,   9,   8,   8,   7,   7,
         7,     7,   7,    7,    6,     4,   3,   0,  -1 );

  NDefaultPriority = 2;

{TRepExprNode}

constructor TRepExprNode.Create(AName: String);
begin
  inherited Create;
  ExprName := AName;
  Left     := nil;
  Right    := nil;
  Line     := -1;
  Col      := -1;
end;

{TRepExprParser}

constructor TRepExprParser.Create(AStreamParser: TStreamTokenizer;
  Reporter: TCustomScriptRep);
begin
  inherited Create;
  FStreamParser := AStreamParser;
  FRootNode := nil;
  FReporter := Reporter;
end;

destructor TRepExprParser.Destroy;
begin
  inherited Destroy;
end;

procedure TRepExprParser.BuildTree(BreakSymbols: String);
begin
  FBreakSymbols := BreakSymbols;
  FRootNode := ParseNextExpr(FRootNode, 0);
end;

function TRepExprParser.ParseNextExpr(ANode: TRepExprNode;
   TokenPriority: Integer): TRepExprNode;
var
  CurTokenPriority: Integer;
  RootNode: TRepExprNode;
  tmp: TRepExprNode;
  TokenText: String;
  BreakFound: Boolean;
begin
  Result := ANode;
  RootNode := nil;
  BreakFound := False;

  while True do
    with FStreamParser do
    begin
      if EOS or BreakFound then
      begin
        if Result <> nil then
          Result.Right := RootNode
        else
          Result := RootNode;
        break;
      end;

      CurTokenPriority := GetTokenPriority(Token);
      TokenText := Token.Text;

      if (CurTokenPriority <= TokenPriority) and
        (CurTokenPriority <> NDefaultPriority) then
      begin
        if Result <> nil then
          Result.Right := RootNode
        else
          Result := RootNode;
        break;
      end;

      if (TokenText <> '(') and (TokenText <> ')') and
        (not TokenBreak(TokenText)) then
      begin
        with Token do
        begin
          tmp      := TRepExprNode.Create(Text);
          tmp.Line := Line;
          tmp.Col  := Col;
        end;

        Next;

        Tmp.Left := RootNode;
        RootNode := Tmp;

        if ((RootNode <> nil) and (RootNode.Left <> nil)) or
          (CurTokenPriority <> NDefaultPriority) then
          RootNode := ParseNextExpr(RootNode, CurTokenPriority);
      end
      else if TokenText = '(' then
      begin
        Next;
        RootNode := ParseNextExpr(RootNode, NDefaultPriority-1);
        with Token do
          if EOS then
            FReporter.ScriptError(67, Line, Col)
          else if Text <> ')' then
            FReporter.ScriptError(68, Line, Col);
        Next;
      end
      else
        BreakFound := TokenBreak(TokenText);
    end;
end;

function TRepExprParser.GetTokenPriority(Token: TToken): Integer;
var
  i: Integer;
begin
  // if token's text will not be found in ExprNameArray, it priority will
  // sets to 3 !!!!

  Result := NDefaultPriority;

  for i := Low(ExprNameArray) to High(ExprNameArray) do
    if ExprNameArray[i] = AnsiUpperCase(Token.Text) then
    begin
      Result := ExprPriorityArray[i];
      break;
    end
end;

function TRepExprParser.TokenBreak(ATokenText: String): Boolean;
var
  p: Integer;
begin

  p := Pos(AnsiUpperCase(ATokenText), AnsiUpperCase(FBreakSymbols));
  Result := (p > 0) and
    (
      (p = 1) or
      ((p > 1) and (FBreakSymbols[p-1] = ' '))
    ) and
    (
      (p + Length(ATokenText)-1 = Length(FBreakSymbols)) or
      ((p < Length(FBreakSymbols)) and (FBreakSymbols[p+Length(ATokenText)] = ' '))
    ) and
    (ATokenText <> ',');

end;


{Mescellenious functions}

procedure UtilizeTree(RootNode: TRepExprNode);
begin
  if RootNode = nil then exit;
  UtilizeTree(RootNode.Left);
  UtilizeTree(RootNode.Right);
  RootNode.Free;
end;

end.
