{**
*
*         Script Reporter Component library
*         Report script debugger
*          Vlad Ilushchenko 1998-99
*
*
**}

unit RepDebugger;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ExtCtrls, ComCtrls, StdCtrls, ScriptRep, RepParamForm, Oracle, RepKrnl, Reporter;

type
  TWaitingType = (wtNone, wtTerminate, wtFirstPage);

  TRepDebugger = class;
  TRepDebuggerNotifyEvent = procedure (Sender: TRepDebugger;
                                       Reporter: TScriptRep) of object;

  TRepDebugger = class(TWinControl)
  private
    { Private declarations }
    FMemo: TMemo;
    FSplitter: TSplitter;
    FListBox: TListBox;
    FSplitPercent: Integer;
    FReporter: TScriptRep;
    FParamForm: TfmRepParam;
    FBeforePrepare: TRepDebuggerNotifyEvent;
    FBeforeRun: TRepDebuggerNotifyEvent;
    FAfterRun: TRepDebuggerNotifyEvent;
    FOnError: TRepDebuggerNotifyEvent;
    procedure SplitterMoved(Sender: TObject);
    procedure ListBoxClick(Sender: TObject);
    function GetSession: TOracleSession;
    procedure SetSession(Value: TOracleSession);
    function GetEditFont: TFont;
    procedure SetEditFont(Value: TFont);
    function GetLines: TStrings;
    procedure SetLines(Value: TStrings);
    function GetModified: Boolean;
    procedure SetModified(Value: Boolean);
    function GetBeforeSubRepParsed: TSubReportParseEvent;
    procedure SetBeforeSubRepParsed(Value: TSubReportParseEvent);

    procedure MarkLine(Line, Col: Integer);

  protected
    { Protected declarations }
    procedure WMSize(var Msg: TWMSize); message WM_SIZE;
    function TextOffset(Line, Col: Integer): Integer;
    function GetPosition(TextOffset: Integer): TRepPosition;
  public
    { Public declarations }
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;

    function RunScript(WaitingType: TWaitingType): Boolean;
    property Modified: Boolean read GetModified write SetModified;

    property Reporter: TScriptRep read FReporter;
  published
    { Published declarations }
    property Session: TOracleSession read GetSession write SetSession;
    property EditFont: TFont read GetEditFont write SetEditFont;
    property Lines: TStrings read GetLines write SetLines;
    property Align;

    property BeforePrepare: TRepDebuggerNotifyEvent read FBeforePrepare write FBeforePrepare;
    property BeforeRun: TRepDebuggerNotifyEvent read FBeforeRun write FBeforeRun;
    property AfterRun: TRepDebuggerNotifyEvent read FAfterRun write FAfterRun;
    property OnError: TRepDebuggerNotifyEvent read FOnError write FOnError;
    property BeforeSubReportParsed: TSubReportParseEvent read GetBeforeSubRepParsed write SetBeforeSubRepParsed;
  end;

procedure Register;

implementation

uses ObjList, StringTools;

const
  STempFile = '$$rep$dbg$tmp$file.$$$';

procedure Register;
begin
  RegisterComponents('ISD', [TRepDebugger]);
end;

constructor TRepDebugger.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FMemo := TMemo.Create(Self);
  FSplitter := TSplitter.Create(Self);
  FListBox := TListBox.Create(Self);
  FReporter := TScriptRep.Create(Self);
  FParamForm := TfmRepParam.Create(Self);

  FSplitPercent := 20;

  with FListBox do
  begin
    Parent := Self;
    Height := Round(Self.Height * FSplitPercent / 100);
    Align  := alBottom;
    OnClick := ListBoxClick;
  end;

  with FSplitter do
  begin
    Parent  := Self;
    Align   := alBottom;
    OnMoved := SplitterMoved;
    Cursor  := crVSplit;
  end;

  with FMemo do
  begin
    Parent := Self;
    Align  := alClient;
    Font.Name := 'COURIER NEW';
    Font.Charset := RUSSIAN_CHARSET;
//    PlainText := True;
    WordWrap  := False;
  end;
end;

destructor TRepDebugger.Destroy;
begin
  if not FReporter.Terminated then FReporter.Terminate;
  inherited Destroy;
end;

procedure TRepDebugger.SplitterMoved(Sender: TObject);
begin
  FSplitPercent := Max(Min(Round((Height - FSplitter.Top) / Height * 100), 95), 5);
end;

procedure TRepDebugger.WMSize(var Msg: TWMSize);
begin
  inherited;

  FListBox.Height := Max(Min(Round(Height * FSplitPercent / 100), Height - 10), 10);
  FSplitter.Top := Height - FListBox.Height;
end;

function TRepDebugger.RunScript(WaitingType: TWaitingType): Boolean;
begin
  Result := False;
  if not FReporter.Terminated then
    if MessageDlg('Reporter works at background. Terminate?', mtWarning,
                  [mbYes, mbNo], 0) = mrYes then
      FReporter.Terminate
    else
      exit;

  try
    try
      Lines.SaveToFile(STempFile);
      with FReporter do
      begin
        ScriptFile := STempFile;

        Prepared  := False;
        Processed := False;

        if FParamForm.OpenForm(FReporter) = mrCancel then
          exit
        else
          Application.ProcessMessages;

        if HasErrors then
          raise EReportTerminated.Create('Parse error', -1);

        MultiThreaded := True;
        if Assigned(FBeforePrepare) then FBeforePrepare(Self, FReporter);

        Prepared := True;

        if Assigned(FBeforeRun) then FBeforeRun(Self, FReporter);

        Processed := True;
        if WaitingType = wtTerminate then
          WaitForTerminate
        else if WaitingType = wtFirstPage then
          WaitForNextPage;

        if HasErrors then
          raise EReportTerminated.Create('Run-time error', -2);

        if Assigned(FAfterRun) then FAfterRun(Self, FReporter);


        FListBox.Items.Clear;
        Result := True;
      end;
    finally
      DeleteFile(STempFile);
    end;
  except
    on EReportTerminated do
    begin
      FListBox.Clear;
      FListBox.Items.Assign(FReporter.ErrorList);
      if Assigned(FOnError) then FOnError(Self, FReporter);
    end;
  end;
end;

function TRepDebugger.GetSession: TOracleSession;
begin
  Result := FReporter.OracleSession;
end;

procedure TRepDebugger.SetSession(Value: TOracleSession);
begin
  FReporter.OracleSession := Value;
end;

function TRepDebugger.TextOffset(Line, Col: Integer): Integer;
var
  i: Integer;
begin
  Result := 0;

  if Line = -1 then exit;

  with FMemo do
    for i := 0 to Line-1 do
      Inc(Result, Length(Lines[i])+2);

  Inc(Result, Col);
end;

procedure TRepDebugger.ListBoxClick(Sender: TObject);
begin
  with FListBox do
    if Items.Objects[ItemIndex] <> nil then
      with TRepPosition(Pointer(Items.Objects[ItemIndex])^) do
        MarkLine(Line, Col);
end;

procedure TRepDebugger.MarkLine(Line, Col: Integer);
begin
  FMemo.SelStart := TextOffset(Line, Col);
  FMemo.SelLength := 0;
  FMemo.SetFocus;
  
  SendMessage(FMemo.Handle, WM_KEYDOWN, VK_RIGHT, 0);
  if FMemo.SelStart <> TextOffset(Line, Col) then
    SendMessage(FMemo.Handle, WM_KEYDOWN, VK_LEFT, 0);

  SendMessage(FMemo.Handle, WM_KEYDOWN, VK_UP, 0);
  if FMemo.SelStart <> TextOffset(Line, Col) then
    SendMessage(FMemo.Handle, WM_KEYDOWN, VK_DOWN, 0);
  FMemo.SelStart := TextOffset(Line, Col);
  FMemo.SelLength := 0;
end;

function TRepDebugger.GetPosition(TextOffset: Integer): TRepPosition;
begin
  Result.Line := 0;
  Result.Col  := 0;
  with FMemo do
    while (Length(Lines[Result.Line])+2) <= TextOffset do
    begin
      Dec(TextOffset, Length(Lines[Result.Line])+2);
      Inc(Result.Line);
    end;

  Result.Col := TextOffset;
end;

function TRepDebugger.GetEditFont: TFont;
begin
  Result := FMemo.Font;
end;

procedure TRepDebugger.SetEditFont(Value: TFont);
begin
  FMemo.Font := Value;
end;

function TRepDebugger.GetLines: TStrings;
begin
  Result := FMemo.Lines;
end;

procedure TRepDebugger.SetLines(Value: TStrings);
begin
  FMemo.Lines := Value;
end;

function TRepDebugger.GetModified: Boolean;
begin
  Result := FMemo.Modified;
end;

procedure TRepDebugger.SetModified(Value: Boolean);
begin
  FMemo.Modified := Value;
end;

procedure TRepDebugger.SetBeforeSubRepParsed;
begin
  FReporter.BeforeSubReportParsed := Value;
end;

function TRepDebugger.GetBeforeSubRepParsed;
begin
  Result := FReporter.BeforeSubReportParsed;
end;

end.
