{**
*
*         Script Reporter Component library
*         Script language extensions
*          Vlad Ilushchenko 1998-99
*
*
**}
unit RepExpr;

interface

uses SysUtils, RepKrnl, ObjList, StringTools, Mask;

type
  {Various kinds of RepExpressions}

  {
    To get the value of this expression you need to make reporter's turn.
    To this reason it can't be initial for other expressions
  }

  TRepFieldExpressionClass = class of TRepFieldExpression;

  TRepFieldExpression = class(TRepExpression)
  private
    FValue: Variant;
    FPrevValue: Variant;
    FField: TRepToken;
    FQuery: TRepToken;
    FRepQuery: TRepQuery;
  protected
    function GetAsVariant: Variant; override;
    procedure Turn; override;
    procedure SetTurnable(Value: Boolean); override;
  public
    constructor Create(RepEngine: TCustomRepEngine; Pos: TRepPosition;
                       Query, Field: TRepToken);
    procedure Prepare; override;
    function GetOldValue: Variant; override;
  end;

  TRepConstExpression = class(TRepExpression)
  private
    FValue: Variant;
  protected
    function GetAsVariant: Variant; override;
    procedure SetAsVariant(AValue: Variant); override;
  public
    constructor Create(RepEngine: TCustomRepEngine; ConstValue: Variant;
      AName: String; Pos: TRepPosition);
  end;

  {Basic classes for plug-in routines}

  TRepUnaryExpressionClass = class of TRepUnaryExpression;
  TRepBinaryExpressionClass = class of TRepBinaryExpression;
  TRepSetExpressionClass = class of TRepSetExpression;

  {
    Description
    -----------

    TRepUnaryExpression descendant MUST override GetAsVariant method to define
    result value.

    procedure Prepare; override this method to get address of needed object. By
                       default this method obtain address of TRepExpression
                       named ArgName
    property ArgName;  name of argument
    property ArgValue; value of argument (call this method only if default
                       Prepare method was called)

    Available values for Arg parameter in call to AddPlugIn:
                                                  ---------
    [akRight] for unary right sided operators like unary -(minus):  -10
                                                                    ^
    or
    [akLeft]  for unary left sided operators like a factorial: 4!
                                                                ^
    or
    [] for operators, or functions without any parameter: PI
                                                          ^

    Note:
    ----
    To write funtion without any parameter you should disallow
    to call to default Prepare method

    TMyExpr = class(TRepUnaryExpression)
      ...
      procedure Prepare; override;
      ...
    end;

    ...

    procedure TMyExpr.Prepare;
    begin
      // do nothing to prevent inherited method to be executed
    end;
  }

  TRepUnaryExpression = class(TRepExpression)
  private
    FValue: Variant;
    FPrevValue: Variant;
    FExpr: TRepToken;
    function GetArg: TRepToken;
  protected
    function GetAsVariant: Variant; override;
    procedure SetAsVariant(AValue: Variant); override;
    function GetOldValue: Variant; override;
    procedure Prepare; override;

    property Arg: TRepToken read GetArg;
  public
    constructor Create(RepEngine: TCustomRepEngine;
      Expr, WhenClause, ResetClause: TRepToken;
      AName: String; Pos: TRepPosition); virtual;
  end;

  {TRepBinaryExpression}

  {
    Available values for Arg parameter in call to AddPlugIn:
                                                  ---------
    [akRight, akLeft] for binary operators: 10 + 4
                                               ^
  }

  TRepBinaryExpression = class(TRepExpression)
  private
    FLeft: TRepToken;
    FRight: TRepToken;
    function GetLeft: TRepToken;
    function GetRight: TRepToken;
  protected
    procedure Prepare; override;
    property Left: TRepToken read GetLeft;
    property Right: TRepToken read GetRight;
  public
    constructor Create(RepEngine: TCustomRepEngine;
                       ALeft, ARight: TRepToken;
                       AName: String; Pos: TRepPosition); virtual;
  end;

  {TRepSetExpression}

  {
    Available values for Arg parameter in call to AddPlugIn:
                                                  ---------
    [akRight, akSet] if your write function such as F(a,b,c)
                                                    ^
    [akLeft, akSet] if your write function such as (a,b,c)F   ;-)
                                                          ^
  }

  TRepSetExpression = class(TRepExpression)
  private
    FValues: TObjectList;
    FArgs: TRepToken;
    function GetValue(Index: Integer): Variant;
    function GetValCount: Integer;
    function GetName(Index: Integer): String;
  protected
    function DefFunction(var AName: String): Integer; dynamic;
    procedure Prepare; override;
  public
    constructor Create(ARepEngine: TCustomRepEngine;
                       Args: TRepToken;
                       AName: String;
                       Pos: TRepPosition); virtual;
    destructor Destroy; override;
    property Values[Index: Integer]: Variant read GetValue;
    property Names[Index: Integer]: String read GetName;
    property ValCount: Integer read GetValCount;
  end;

  {Descendant classes}

  TRepAccumulativeSum = class(TRepUnaryExpression)
  protected
    procedure Turn; override;
    procedure Prepare; override;
  public
    constructor Create(RepEngine: TCustomRepEngine;
      Expr, WhenClause, ResetClause: TRepToken;
      AName: String; Pos: TRepPosition); override;
  end;

  TRepAccumulativeMin = class(TRepUnaryExpression)
  protected
    procedure Turn; override;
    procedure Prepare; override;
  public
    constructor Create(RepEngine: TCustomRepEngine;
      Expr, WhenClause, ResetClause: TRepToken;
      AName: String; Pos: TRepPosition); override;
  end;

  TRepAccumulativeMax = class(TRepUnaryExpression)
  protected
    procedure Turn; override;
    procedure Prepare; override;
  public
    constructor Create(RepEngine: TCustomRepEngine;
      Expr, WhenClause, ResetClause: TRepToken;
      AName: String; Pos: TRepPosition); override;
  end;

  TRepAccumulativeAvg = class(TRepUnaryExpression)
  private
    FExprCount: Integer;
  protected
    function GetAsVariant: Variant; override;
    procedure Prepare; override;
    procedure Turn; override;
  public
    constructor Create(RepEngine: TCustomRepEngine;
      Expr, WhenClause, ResetClause: TRepToken;
      AName: String; Pos: TRepPosition); override;
    function Reset: Boolean; override;
  end;

  TRepCheckExpression = class(TRepUnaryExpression)
  private
    FPrevValue: Variant;
  protected
    function GetAsVariant: Variant; override;
    procedure Turn; override;
    procedure Prepare; override;
  public
    constructor Create(RepEngine: TCustomRepEngine;
      Expr, WhenClause, ResetClause: TRepToken;
      AName: String; Pos: TRepPosition); override;
  end;

  TRepBinPlusExpr = class(TRepBinaryExpression)
  protected
    function GetAsVariant: Variant; override;
  end;

  TRepBinMinusExpr = class(TRepBinaryExpression)
  protected
    function GetAsVariant: Variant; override;
  end;

  TRepBinMulExpr = class(TRepBinaryExpression)
  protected
    function GetAsVariant: Variant; override;
  end;

  TRepBinDivExpr = class(TRepBinaryExpression)
  protected
    function GetAsVariant: Variant; override;
  end;

  TRepBinAndExpr = class(TRepBinaryExpression)
  protected
    function GetAsVariant: Variant; override;
  end;

  TRepBinOrExpr = class(TRepBinaryExpression)
  protected
    function GetAsVariant: Variant; override;
  end;

  TRepBinXorExpr = class(TRepBinaryExpression)
  protected
    function GetAsVariant: Variant; override;
  end;

  TRepUnaryMinusExpr = class(TRepUnaryExpression)
  protected
    function GetAsVariant: Variant; override;
  end;

  TRepNotExpr = class(TRepUnaryExpression)
  protected
    function GetAsVariant: Variant; override;
  end;

  TRepMoreExpr = class(TRepBinaryExpression)
  protected
    function GetAsVariant: Variant; override;
  end;

  TRepMoreOrEquelExpr = class(TRepBinaryExpression)
  protected
    function GetAsVariant: Variant; override;
  end;

  TRepLessExpr = class(TRepBinaryExpression)
  protected
    function GetAsVariant: Variant; override;
  end;

  TRepLessOrEquelExpr = class(TRepBinaryExpression)
  protected
    function GetAsVariant: Variant; override;
  end;

  TRepNotEquelExpr = class(TRepBinaryExpression)
  protected
    function GetAsVariant: Variant; override;
  end;

  TRepEquelExpr = class(TRepBinaryExpression)
  protected
    function GetAsVariant: Variant; override;
  end;

  TRepDuplicateExpr = class(TRepUnaryExpression)
  protected
    function GetAsVariant: Variant; override;
    function GetOldValue: Variant; override;
    procedure SetAsVariant(Value: Variant); override;
    procedure SetTurnable(Value: Boolean); override;
    procedure Prepare; override;
    procedure Turn; override;
    procedure EndTurn; override;
  end;

  TRepPrevFiledValExpr = class(TRepUnaryExpression)
  protected
    function GetAsVariant: Variant; override;
    procedure Prepare; override;
    procedure Turn; override;
    procedure EndTurn; override;
  public
    constructor Create(RepEngine: TCustomRepEngine;
      Expr, WhenClause, ResetClause: TRepToken;
      AName: String; Pos: TRepPosition); override;
  end;

  TRepCaseExpression = class(TRepSetExpression)
  protected
    function DefFunction(Var AName: String): integer; override;
    function GetAsVariant: Variant; override;
  end;

  TRepNvlExpression = class(TRepSetExpression)
  protected
    function DefFunction(Var AName: String): integer; override;
    function GetAsVariant: Variant; override;
  end;

  TRepQueryOpenedExpr = class(TRepUnaryExpression)
  private
    FQuery: TRepQuery;
  protected
    function GetAsVariant: Variant; override;
    procedure Prepare; override;
  end;

  TRepRowsFetchedCountExpr = class(TRepUnaryExpression)
  private
    FQuery: TRepQuery;
  protected
    function GetAsVariant: Variant; override;
    procedure Prepare; override;
  end;

  TRepEOFExpression = class(TRepUnaryExpression)
  private
    FQuery: TRepQuery;
  protected
    function GetAsVariant: Variant; override;
    procedure Prepare; override;
  end;

  TEmptyQueryExpression = class(TRepUnaryExpression)
  private
    FQuery: TRepQuery;
  protected
    function GetAsVariant: Variant; override;
    procedure Prepare; override;
  end;

  TRepErrorSubstExpr = class(TRepSetExpression)
  protected
    function DefFunction(Var AName: String): integer; override;
    function GetAsVariant: Variant; override;
  end;

  TRepDebugExpr = class(TRepSetExpression)
  protected
    function DefFunction(Var AName: String): integer; override;
    function GetAsVariant: Variant; override;
  end;

  TRepLeftStrExpr = class(TRepSetExpression)
  protected
    function DefFunction(Var AName: String): integer; override;
    function GetAsVariant: Variant; override;
  end;

  TRepRightStrExpr = class(TRepSetExpression)
  protected
    function DefFunction(Var AName: String): integer; override;
    function GetAsVariant: Variant; override;
  end;

  TRepSubStrExpr = class(TRepSetExpression)
  protected
    function DefFunction(Var AName: String): integer; override;
    function GetAsVariant: Variant; override;
  end;

  TRepAssignExpr = class(TRepBinaryExpression)
  protected
    function GetAsVariant: Variant; override;
  end;

  TRepReplicateExpr = class(TRepSetExpression)
  protected
    function DefFunction(Var AName: String): integer; override;
    function GetAsVariant: Variant; override;
  end;

  TRepFormatExpr = class(TRepSetExpression)
  protected
    function DefFunction(Var AName: String): integer; override;
    function GetAsVariant: Variant; override;
  end;

  TRepCtrlProcessedExpr = class(TRepUnaryExpression)
  private
    FControl: TRepControl;
  protected
    function GetAsVariant: Variant; override;
    procedure Prepare; override;
  end;

  TRepQueryProcessedExpr = class(TRepUnaryExpression)
  private
    FQuery: TRepQuery;
  protected
    function GetAsVariant: Variant; override;
    procedure Prepare; override;
  end;

  TRepChrExpression = class(TRepSetExpression)
  protected
    function DefFunction(Var AName: String): integer; override;
    function GetAsVariant: Variant; override;
  end;

  TSubRepExpression = class(TRepBinaryExpression)
  private
    FWorkRepEngine: TCustomRepEngine;
  protected
    function GetAsVariant: Variant; override;
    procedure Prepare; override;
  end;

  TSubRepProcessedExpr = class(TRepUnaryExpression)
  private
    FReport: TCustomRepEngine;
  protected
    function GetAsVariant: Variant; override;
    procedure Prepare; override;
  end;

  TRepDecodeExpression = class(TRepSetExpression)
  protected
    procedure Prepare; override;
    function GetAsVariant: Variant; override;
  end;

  TRepPosExpression = class(TRepSetExpression)
  protected
    function DefFunction(Var AName: string): Integer; override;
    function GetAsVariant: Variant; override;
  end;

  TRepToDateExpression = class(TRepSetExpression)
  protected
    function DefFunction(Var AName: string): Integer; override;
    function GetAsVariant: Variant; override;
  end;

  TRepToNumberExpression = class(TRepSetExpression)
  protected
    function DefFunction(Var AName: string): Integer; override;
    function GetAsVariant: Variant; override;
  end;

  TRepToStringExpression = class(TRepSetExpression)
  protected
    function DefFunction(Var AName: string): Integer; override;
    function GetAsVariant: Variant; override;
  end;

  TRepPageHeightExpression = class(TRepUnaryExpression)
  protected
    procedure Prepare; override;
    function GetAsVariant: Variant; override;
  end;

  TRepExtraHeightExpression = class(TRepUnaryExpression)
  private
    FPInt: ^Integer;
  protected
    procedure Prepare; override;
    function GetAsVariant: Variant; override;
  end;

  TRepExtraHeightUsedExpression = class(TRepUnaryExpression)
  private
    FPInt: ^Integer;
  protected
    procedure Prepare; override;
    function GetAsVariant: Variant; override;
  end;

  TRepCurPageHeightExpression = class(TRepUnaryExpression)
  private
    FPInt: ^Integer;
  protected
    procedure Prepare; override;
    function GetAsVariant: Variant; override;
  end;

  TRepCtrlHeightExpr = class(TRepUnaryExpression)
  private
    FControl: TRepControl;
  protected
    function GetAsVariant: Variant; override;
    procedure Prepare; override;
  end;

  TRepRoundExpression = class(TRepSetExpression)
  protected
    function DefFunction(Var AName: String): integer; override;
    function GetAsVariant: Variant; override;
  end;

  TRepTruncExpression = class(TRepSetExpression)
  protected
    function DefFunction(Var AName: String): integer; override;
    function GetAsVariant: Variant; override;
  end;

  TRepLengthExpression = class(TRepSetExpression)
  protected
    function DefFunction(Var AName: String): integer; override;
    function GetAsVariant: Variant; override;
  end;

  TRepLTrimExpression = class(TRepSetExpression)
  protected
    function DefFunction(Var AName: String): integer; override;
    function GetAsVariant: Variant; override;
  end;

  TRepRTrimExpression = class(TRepSetExpression)
  protected
    function DefFunction(Var AName: String): integer; override;
    function GetAsVariant: Variant; override;
  end;

  TRepTrimExpression = class(TRepSetExpression)
  protected
    function DefFunction(Var AName: String): integer; override;
    function GetAsVariant: Variant; override;
  end;

  TRepLocalDate = class(TRepUnaryExpression)
  protected
    function GetAsVariant: Variant; override;
    procedure Prepare; override;
  end;

  TRepLocalTime = class(TRepUnaryExpression)
  protected
    function GetAsVariant: Variant; override;
    procedure Prepare; override;
  end;

  TRepNull = class(TRepUnaryExpression)
  protected
    function GetAsVariant: Variant; override;
    procedure Prepare; override;
  end;

implementation

uses Controls, Forms, Windows, Math, Reporter;

{TRepFieldExpression}

constructor TRepFieldExpression.Create(RepEngine: TCustomRepEngine; Pos: TRepPosition;
                                       Query, Field: TRepToken);
begin
  inherited CreateOriginal(RepEngine,
                           EmptyRepToken,
                           EmptyRepToken,
                           Query.Text+'_'+Field.Text,
                           Pos);
  FQuery := Query;
  FField := Field;
end;

function TRepFieldExpression.GetOldValue: Variant;
begin
  Result := FPrevValue;
end;

function TRepFieldExpression.GetAsVariant: Variant;
begin
  FRepQuery.MakeTurn;
  if Turnable and (VarType(FValue) <> varEmpty) then
  begin
    MakeTurn;
    Result := FValue
  end
  else
    if not FRepQuery.NullFields then
      Result := FRepQuery.Fields[FField.Text]
    else
      Result := null;
end;

procedure TRepFieldExpression.Turn;
begin
  FPrevValue := FValue;

  if not FRepQuery.NullFields then
    FValue := FRepQuery.Fields[FField.Text]
  else
    FValue := null;
end;

procedure TRepFieldExpression.SetTurnable(Value: Boolean);
begin
  inherited SetTurnable(Value);
end;

procedure TRepFieldExpression.Prepare;
begin
  if SelfPrepared then exit;
  inherited Prepare;
  FRepQuery := RepEngine.QueryByToken[FQuery];
  SelfPrepared := True;
end;

{TRepConstExpression}

constructor TRepConstExpression.Create(RepEngine: TCustomRepEngine;
  ConstValue: Variant; AName: String; Pos: TRepPosition);
var
  S: String;
begin
  inherited CreateOriginal(RepEngine,
                           EmptyRepToken,
                           EmptyRepToken,
                           AName,
                           Pos);
  FValue := ConstValue;
  if VarType(ConstValue) = varString then
  begin
    S := FValue;
    if (Length(S) > 0) and (S[1] = '''') then
      FValue := Copy(S, 2, Length(S)-2);
  end;
end;

function TRepConstExpression.GetAsVariant: Variant;
begin
  Result := FValue;
end;

procedure TRepConstExpression.SetAsVariant(AValue: Variant);
begin
  FValue := AValue;
end;

{TRepUnaryExpression}

constructor TRepUnaryExpression.Create(RepEngine: TCustomRepEngine;
  Expr, WhenClause, ResetClause: TRepToken;
  AName: String; Pos: TRepPosition);
begin
  inherited CreateOriginal(RepEngine,
                           WhenClause,
                           ResetClause,
                           AName,
                           Pos);
  FExpr := Expr;
end;

function TRepUnaryExpression.GetOldValue: Variant;
begin
  Result := FPrevValue;
end;

function TRepUnaryExpression.GetAsVariant: Variant;
begin
  Result := FValue;
end;

procedure TRepUnaryExpression.SetAsVariant(AValue: Variant);
begin
  FValue := AValue;
end;

function TRepUnaryExpression.GetArg: TRepToken;
begin
  Result := FExpr;
end;

procedure TRepUnaryExpression.Prepare;
begin
  if SelfPrepared then exit;

  inherited Prepare;

  if FExpr.Addr = nil then
    FExpr.Addr := RepEngine.ExprByToken[FExpr];
  SelfPrepared := True;
end;

{TRepAccumulativeSum}

constructor TRepAccumulativeSum.Create;
begin
  inherited Create(RepEngine, Expr, WhenClause, ResetClause, AName, Pos);
  Turnable := True;
end;

procedure TRepAccumulativeSum.Turn;
begin
  FPrevValue := FValue;
  if FExpr.Addr.Value <> null then
    Inc(FValue, FExpr.Addr.Value);
end;

procedure TRepAccumulativeSum.Prepare;
begin
  if SelfPrepared then exit;
  inherited Prepare;
  FValue := 0;
  SelfPrepared := True;
end;

{TRepAccumulativeMin}

constructor TRepAccumulativeMin.Create;
begin
  inherited Create(RepEngine, Expr, WhenClause, ResetClause, AName, Pos);
  Turnable := True;
end;

procedure TRepAccumulativeMin.Turn;
begin
  FPrevValue := FValue;
  if FExpr.Addr.Value <> null then
    FValue := Min(FValue, FExpr.Addr.Value);
end;

procedure TRepAccumulativeMin.Prepare;
begin
  if SelfPrepared then exit;
  inherited Prepare;
  FValue := $0fffffff;
  SelfPrepared := True;
end;

{TRepAccumulativeMax}

constructor TRepAccumulativeMax.Create;
begin
  inherited Create(RepEngine, Expr, WhenClause, ResetClause, AName, Pos);
  Turnable := True;
end;

procedure TRepAccumulativeMax.Turn;
begin
  FPrevValue := FValue;
  if FExpr.Addr.Value <> null then
    FValue := Max(FValue, FExpr.Addr.Value);
end;

procedure TRepAccumulativeMax.Prepare;
begin
  if SelfPrepared then exit;
  inherited Prepare;
  FValue := $ffffffff;
  SelfPrepared := True;
end;

{TRepAccumulativeAvg}

constructor TRepAccumulativeAvg.Create;
begin
  inherited Create(RepEngine, Expr, WhenClause, ResetClause, AName, Pos);
  Turnable := True;
end;

function TRepAccumulativeAvg.GetAsVariant: Variant;
begin
  Result := inherited Value / FExprCount;
end;

procedure TRepAccumulativeAvg.Prepare;
begin
  if SelfPrepared then exit;
  inherited Prepare;
  FExprCount := 0;
  FValue := 0;
  SelfPrepared := True;
end;

procedure TRepAccumulativeAvg.Turn;
begin
  FPrevValue := FExprCount;
  if FExpr.Addr.Value <> null then
    Inc(FExprCount);
end;

function TRepAccumulativeAvg.Reset;
begin
  Result := inherited Reset;
  FExprCount := 0;
end;

{TRepCheckExpression}

constructor TRepCheckExpression.Create(RepEngine: TCustomRepEngine;
  Expr, WhenClause, ResetClause: TRepToken; AName: String; Pos: TRepPosition);
begin
  inherited Create(RepEngine, Expr, WhenClause, ResetClause, AName, Pos);
  Turnable := True;
end;

procedure TRepCheckExpression.Turn;
begin
  FPrevValue := FValue;
  FValue := FExpr.Addr.Value;
end;

function TRepCheckExpression.GetAsVariant: Variant;
begin
  MakeTurn;
  Result := False;
  if not RepEngine.RepStarted then
    if not RepEngine.RepFinished then
      Result := FPrevValue <> FValue
    else
      Result := True;
end;

procedure TRepCheckExpression.Prepare;
begin
  if SelfPrepared then exit;
  inherited Prepare;
  SelfPrepared := True;
end;

{TRepBinaryExpression}

constructor TRepBinaryExpression.Create(RepEngine: TCustomRepEngine;
                                        ALeft, ARight: TRepToken;
                                        AName: String; Pos: TRepPosition);
begin
  inherited CreateOriginal(RepEngine, EmptyRepToken, EmptyRepToken, AName, Pos);
  FLeft  := ALeft;
  FRight := ARight;
end;

procedure TRepBinaryExpression.Prepare;
begin
  if SelfPrepared then exit;
  with RepEngine do
  begin
    if FLeft.Addr = nil then
      FLeft.Addr := ExprByToken[FLeft];

    if FRight.Addr = nil then
      FRight.Addr := ExprByToken[FRight];
  end;
  SelfPrepared := True;
end;

function TRepBinaryExpression.GetLeft: TRepToken;
begin
  Result := FLeft;
end;

function TRepBinaryExpression.GetRight: TRepToken;
begin
  Result := FRight;
end;

{TRepBinPlusExpr}

function TRepBinPlusExpr.GetAsVariant: Variant;
begin
  Result := Left.Addr.Value + Right.Addr.Value;
end;

{TRepBinMinusExpr}

function TRepBinMinusExpr.GetAsVariant: Variant;
begin
  Result := Left.Addr.Value - Right.Addr.Value;
end;

{TRepBinMulExpr}

function TRepBinMulExpr.GetAsVariant: Variant;
begin
  Result := Left.Addr.Value * Right.Addr.Value;
end;

{TRepBinDivExpr}

function TRepBinDivExpr.GetAsVariant: Variant;
begin
  Result := Left.Addr.Value / Right.Addr.Value;
end;

{TRepBinAndExpr}

function TRepBinAndExpr.GetAsVariant: Variant;
begin
  Result := Left.Addr.Value;
  if not Result then
    exit
  else
    Result := Result and Right.Addr.Value;
end;


{TRepBinOrExpr}

function TRepBinOrExpr.GetAsVariant: Variant;
begin
  Result := Left.Addr.Value or Right.Addr.Value;
end;

{TRepBinXorExpr}

function TRepBinXorExpr.GetAsVariant: Variant;
begin
  Result := Left.Addr.Value xor Right.Addr.Value;
end;

{TRepUnaryMinusExpr}

function TRepUnaryMinusExpr.GetAsVariant: Variant;
begin
  Result := -FExpr.Addr.Value;
end;

{TRepNotExpr}

function TRepNotExpr.GetAsVariant: Variant;
begin
  Result := not FExpr.Addr.Value;
end;

{TRepMoreExpr}

function TRepMoreExpr.GetAsVariant: Variant;
begin
  Result := Left.Addr.Value > Right.Addr.Value;
end;

{TRepMoreOrEquelExpr}

function TRepMoreOrEquelExpr.GetAsVariant: Variant;
begin
  Result := Left.Addr.Value >= Right.Addr.Value;
end;

{TRepLessExpr}

function TRepLessExpr.GetAsVariant: Variant;
begin
  Result := Left.Addr.Value < Right.Addr.Value;
end;

{TRepLessOrEquelExpr}

function TRepLessOrEquelExpr.GetAsVariant: Variant;
begin
  Result := Left.Addr.Value <= Right.Addr.Value;
end;

{TRepNotEquelExpr}

function TRepNotEquelExpr.GetAsVariant: Variant;
begin
  Result := Left.Addr.Value <> Right.Addr.Value;
end;

{TRepEquelExpr}

function TRepEquelExpr.GetAsVariant: Variant;
begin
  Result := Left.Addr.Value = Right.Addr.Value;
end;

{TRepDuplicateExpr}

function TRepDuplicateExpr.GetAsVariant: Variant;
begin
  Result := FExpr.Addr.Value;
end;

function TRepDuplicateExpr.GetOldValue: variant;
begin
  Result := FExpr.Addr.OldValue;
end;

procedure TRepDuplicateExpr.SetAsVariant(Value: Variant);
begin
  if (FExpr.Addr <> nil) and (FExpr.Addr is TRepConstExpression) then
    FExpr.Addr.Value := Value;
end;

procedure TRepDuplicateExpr.SetTurnable(Value: Boolean);
begin
  FExpr.Addr.Turnable := Value;
end;

procedure TRepDuplicateExpr.Prepare;
begin
  if SelfPrepared then exit;
  inherited Prepare;

  FExpr.Addr.WhenClause := WhenClause;
  FExpr.Addr.ResetClause := ResetClause;

  FExpr.Addr.ResetBefore := ResetBefore;
  FExpr.Addr.ResetAfter  := ResetAfter;

  with FExpr.Addr do Prepare;

  SelfPrepared := True;
end;

procedure TRepDuplicateExpr.Turn;
begin
  FExpr.Addr.MakeTurn;
end;

procedure TRepDuplicateExpr.EndTurn;
begin
  FExpr.Addr.FinishTurn;
  inherited EndTurn;
end;

{TRepPrevFiledValExpr}

constructor TRepPrevFiledValExpr.Create(RepEngine: TCustomRepEngine;
      Expr, WhenClause, ResetClause: TRepToken; AName: String; Pos: TRepPosition);
begin
  inherited Create(RepEngine, Expr, WhenClause, ResetClause, AName, Pos);
  Turnable := True;
end;

function TRepPrevFiledValExpr.GetAsVariant: Variant;
begin
  Result := FExpr.Addr.OldValue
end;

procedure TRepPrevFiledValExpr.Prepare;
begin
  if SelfPrepared then exit;
  inherited Prepare;
  FExpr.Addr.Bind;
  SelfPrepared := True;
end;

procedure TRepPrevFiledValExpr.Turn;
begin
  FExpr.Addr.MakeTurn;
end;

procedure TRepPrevFiledValExpr.EndTurn;
begin
  inherited EndTurn;
  FExpr.Addr.FinishTurn;
end;

{TRepSetExpression}

constructor TRepSetExpression.Create(ARepEngine: TCustomRepEngine;
                                     Args: TRepToken; AName: String; Pos: TRepPosition);
begin
  inherited CreateOriginal(ARepEngine, EmptyRepToken, EmptyRepToken, aName, Pos);
  FValues    := TObjectList.Create;
  FArgs      := Args;
end;

destructor TRepSetExpression.Destroy;
begin
  FValues.Clear;
  FValues.Free;
  inherited Destroy;
end;

function TRepSetExpression.GetValue(Index: Integer): Variant;
begin
  Result := TRepExpression(FValues.Items[Index]).Value;
end;

function TRepSetExpression.GetName(Index: Integer): String;
begin
  Result := TRepExpression(FValues.Items[Index]).Name;
end;

procedure TRepSetExpression.Prepare;
var
  ExprName: String;
  ExprAddr: String;
  ArgS: String;
  AddrS: String;
  CommaPosArg: Integer;
  CommaPosAddr: Integer;
  FuncName: String;
  ArgCount: Integer;
begin
  if SelfPrepared then exit;

  inherited Prepare;

  ArgS := FArgs.Text;
  AddrS := FArgs.AddrStr;
  FValues.Clear;

  while True do
  begin
    if Trim(ArgS) = EmptyStr then break;

    CommaPosArg  := Pos(',', ArgS);
    CommaPosAddr := Pos(',', AddrS);

    if CommaPosArg > 0 then
    begin
      ExprName := Copy(ArgS, 1, CommaPosArg - 1);
      ExprAddr := Copy(AddrS, 1, CommaPosAddr - 1);

      ArgS := Copy(ArgS, CommaPosArg + 1, Length(ArgS));
      AddrS := Copy(AddrS, CommaPosAddr + 1, Length(AddrS));
    end
    else
    begin
      ExprName := ArgS;
      ExprAddr := AddrS;
      ArgS  := EmptyStr;
      AddrS := EmptyStr;
    end;

    if ExprAddr = '0' then
      FValues.Add(RepEngine.ExprByToken[ToRepToken(ExprName, nil, '', RepPosition(-1,-1))])
    else
      FValues.Add(Pointer(StrToInt(ExprAddr)));
  end;

  ArgCount := DefFunction(FuncName);
  if FuncName <> 'UNDEFINED' then
    if ValCount < ArgCount then
      RepEngine.ReportError(17, FuncName, Self.Name, Self, True)
    else if ArgCount < ValCount then
      RepEngine.ReportError(18, FuncName, Self.Name, Self, True);

  SelfPrepared := True;

end;

function TRepSetExpression.DefFunction(Var AName: String): Integer;
begin
  AName := 'UNDEFINED';
  Result := 0;
end;

function TRepSetExpression.GetValCount: Integer;
begin
  Result := FValues.Count;
end;

{TRepCaseExpression}

function TRepCaseExpression.GetAsVariant: Variant;
begin
  if Values[0] = True then
    Result := Values[1]
  else
    Result := Values[2];
end;

function TRepCaseExpression.DefFunction(var AName: string): Integer;
begin
  AName := 'IF';
  Result := 3;
end;

{TRepNvlExpression}

function TRepNvlExpression.GetAsVariant: Variant;
begin
  if (VarType(Values[0]) = varEmpty) or (VarType(Values[0]) = varNull) then
    Result := Values[1]
  else
    Result := Values[2];
end;

function TRepNvlExpression.DefFunction(Var AName: String): Integer;
begin
  AName := 'NVL';
  Result := 3;
end;

{TRepQueryOpenedExpr}

procedure TRepQueryOpenedExpr.Prepare;
begin
  if SelfPrepared then exit;
  FQuery := RepEngine.QueryByToken[FExpr];
  SelfPrepared := True;
end;

function TRepQueryOpenedExpr.GetAsVariant: Variant;
begin
  Result := not FQuery.Executed;
  FQuery.NullFields := False;
end;

{TRepRowsFetchedCountExpr}

procedure TRepRowsFetchedCountExpr.Prepare;
begin
  if SelfPrepared then exit;
  FQuery := RepEngine.QueryByToken[FExpr];
  SelfPrepared := True;
end;

function TRepRowsFetchedCountExpr.GetAsVariant: Variant;
begin
  Result := FQuery.RowsFetched;
end;

{TRepEOFExpression}

procedure TRepEOFExpression.Prepare;
begin
  if SelfPrepared then exit;
  FQuery := RepEngine.QueryByToken[FExpr];
  SelfPrepared := True;
end;

function TRepEOFExpression.GetAsVariant: Variant;
begin
  Result := FQuery.EOF;
end;

{TEmptyQueryExpression}

procedure TEmptyQueryExpression.Prepare;
begin
  if SelfPrepared then exit;
  FQuery := RepEngine.QueryByToken[FExpr];
  SelfPrepared := True;
end;

function TEmptyQueryExpression.GetAsVariant: Variant;
begin
  Result := FQuery.NullFields;
end;

{TRepErrorSubstExpr}

function TRepErrorSubstExpr.GetAsVariant: Variant;
begin
  try
    Result := Values[0];
  except
    Result := Values[1];
  end;
end;

function TRepErrorSubstExpr.DefFunction(Var AName: String): Integer;
begin
  AName := 'ERRORSUBST';
  Result := 2;
end;

{TRepDebugExpr}

function TRepDebugExpr.GetAsVariant: Variant;
var
  s: String;
begin
  Result := Values[1];
  if VarType(Result) = VarBoolean then
    if Result = True then
      s := 'True'
    else
      s := 'False'
  else
    if VarType(Result) = VarNull then
      s := 'NULL'
  else
    s := Result;
  if Application.MessageBox(PChar(Format(Values[0], [s])), PChar(Application.Title),
    MB_ICONHAND or MB_OKCANCEL) = ID_CANCEL then Abort;
end;

function TRepDebugExpr.DefFunction(Var AName: String): Integer;
begin
  AName := 'DEBUG';
  Result := 2;
end;


{TRepLeftStrExpr}

function TRepLeftStrExpr.GetAsVariant: Variant;
begin
  Result := Copy(Values[0], 1, StrToInt(Values[1]));
end;

function TRepLeftStrExpr.DefFunction(Var AName: String): Integer;
begin
  AName := 'LEFTSTR';
  Result := 2;
end;

{TRepRightStrExpr}

function TRepRightStrExpr.GetAsVariant: Variant;
var
  l: Integer;
begin
  l := Length(Values[0]);
  Result := Copy(Values[0], l-StrToInt(Values[1]) + 1, l);
end;

function TRepRightStrExpr.DefFunction(Var AName: String): Integer;
begin
  AName := 'RIGHTSTR';
  Result := 2;
end;

{TRepSubStrExpr}

function TRepSubStrExpr.GetAsVariant: Variant;
begin
  if VarType(Values[0]) = varNull then
    Result := null
  else
    Result := Copy(Values[0], StrToInt(Values[1]), StrToInt(Values[2]));
end;

function TRepSubStrExpr.DefFunction(Var AName: String): Integer;
begin
  AName := 'SUBSTR';
  Result := 3;
end;

{TRepAssignExpr}

function TRepAssignExpr.GetAsVariant: Variant;
begin
  FLeft.Addr.Value := FRight.Addr.Value;
  Result := True;
end;

{TRepReplicateExpr}

function TRepReplicateExpr.GetAsVariant: Variant;
var
  c: Char;
  Count: Integer;
begin
  c := VarToStr(Values[1])[1];
  Count := Values[0];
  Result := ReplicateChar(c, Count);
end;

function TRepReplicateExpr.DefFunction(Var AName: String): Integer;
begin
  AName := 'REPLICATE';
  Result := 2;
end;

{TRepFormatExpr}

function TRepFormatExpr.GetAsVariant: Variant;
var
 aPicture,s: String;
 aWidth, PosDec, Precision: Integer;
 aTrim,aCenter: Boolean;
begin
  aPicture := Values[0];
  aWidth := length(aPicture);
  aTrim := False;
  aCenter := False;

  if Pos('@', aPicture) > 0  then
  begin
    aCenter := True;
    aPicture := copy(aPicture,2,aWidth);
  end;

  if Pos('~',aPicture) > 0 then
    RepEngine.ReportError(19, Name, Self.ClassName, Self, True);

  if (aWidth > 0)then
  begin
   if (aPicture[1] = '^') then
   begin
      aPicture := copy(aPicture,2,aWidth);
      aTrim := True;
    end;
    if AnsiUpperCase(aPicture[1])='W' then
    begin
      aPicture := copy(aPicture,2,aWidth);
      Dec(aWidth);
      if (aWidth > 0)and(aPicture[1]>='0') and (aPicture[1]<='9') then
          aPicture := ReplicateChar('A',StrToInt(aPicture));
    end;
  end;

  PosDec := Pos('.',aPicture);

  if  PosDec > 0 then
    Precision := length(trim(copy(aPicture, PosDec+1,20)))
  else
    Precision := 0;

  case VarType(Values[1]) of
    varSmallint,
    varInteger,
    varCurrency,
    varSingle,
    varDouble:begin
                s := FormatFloat(aPicture,Values[1]);
                s := s+Space(Precision-
                       length(copy(s,Pos(DecimalSeparator,s)+1,20)));
                Result := Padl(s,aWidth);
              end;
    varDate: Result := FormatDateTime(aPicture,Values[1]);
    varBoolean: Result := String(Values[1]);
    varString, varOleStr: begin
                            Result := String(Values[1]);
                            Result := FormatMaskText(aPicture,Result);
                            if aTrim then Result := trim(Result);
                            if aCenter then Result := CenterStr(trim(Result),aWidth);
                          end;
  else
    Result := Space(aWidth);
  end;

end;

function TRepFormatExpr.DefFunction(Var AName: String): Integer;
begin
  AName := 'FORMAT';
  Result := 2;
end;


{TRepCtrlProcessedExpr}

procedure TRepCtrlProcessedExpr.Prepare;
begin
  if SelfPrepared then exit;
  FControl := RepEngine.ControlByToken[FExpr];
  SelfPrepared := True;
end;

function TRepCtrlProcessedExpr.GetAsVariant: Variant;
begin
  Result := FControl.Processed;
end;

{TRepQueryProcessedExpr}

procedure TRepQueryProcessedExpr.Prepare;
begin
  if SelfPrepared then exit;
  FQuery := RepEngine.QueryByToken[FExpr];
  SelfPrepared := True;
end;

function TRepQueryProcessedExpr.GetAsVariant: Variant;
begin
  Result := FQuery.Processed;
end;

{TRepChrExpression}

function TRepChrExpression.GetAsVariant: Variant;
var
  Val: Integer;
begin
  Val := Values[0];
  Result := Chr(Val);
end;

function TRepChrExpression.DefFunction(Var AName: String): Integer;
begin
  AName := 'CHR';
  Result := 1;
end;

{TSubRepExpression}

function TSubRepExpression.GetAsVariant;
begin
  if FWorkRepEngine <> nil then
    FWorkRepEngine.WaitForTerminate;
  Result := Left.Addr.Value;
end;

procedure TSubRepExpression.Prepare;
begin
  if SelfPrepared then exit;
  if AnsiUpperCase(FLeft.Text) = 'SUPER' then
    if RepEngine.ParentReport = nil then
      RepEngine.ReportError(20, Self.Name, '', Self, True)
    else if RepEngine.NoWait and RepEngine.MultiThreaded then
      RepEngine.ReportError(24, Self.Name, '', Self, True)
    else
    begin
      FWorkRepEngine := nil;
      RepEngine.PreparedObject := Self;
      FLeft.Addr := RepEngine.ParentReport.ExprByToken[FRight];
    end
  else
  begin
    FWorkRepEngine := RepEngine.ChildRepName[FLeft];
    FWorkRepEngine.PreparedObject := Self;
    FLeft.Addr := FWorkRepEngine.ExprByToken[FRight];
  end;
end;

{TSubRepProcessedExpr}

function TSubRepProcessedExpr.GetAsVariant: Variant;
begin
  Result := FReport.RepProcessed;
end;

procedure TSubRepProcessedExpr.Prepare;
begin
  if SelfPrepared then exit;
  FReport := RepEngine.ChildRepName[FExpr];
  SelfPrepared := True;
end;

{TRepDecodeExpression}

procedure TRepDecodeExpression.Prepare;
begin
  if SelfPrepared then exit;
  inherited Prepare;
  if ValCount < 3 then
    RepEngine.ReportError(21, Name, Self.ClassName, Self, True);

  if (ValCount mod 2) = 1 then
    RepEngine.ReportError(22, Name, Self.ClassName, Self, True);
  SelfPrepared := True;
end;

function TRepDecodeExpression.GetAsVariant: Variant;
var
  TestVal: Variant;
  i: Integer;
  Cnt: Integer;
begin
  TestVal := Values[0];
  Cnt     := ValCount;

  i := 1;
  while i < Cnt do
  begin
    if i = Cnt-1 then
      Result := Values[i]
    else
      if Values[i] = TestVal then
      begin
        Result := Values[i+1];
        break;
      end;
    Inc(i, 2);
  end
end;

{TRepPosExpression}

function TRepPosExpression.DefFunction(var AName: String): Integer;
begin
  AName := 'POS';
  Result := 2;
end;

function TRepPosExpression.GetAsVariant: Variant;
var
  S1, S2: String;
begin
  S1 := Values[0];
  S2 := Values[1];
  Result := Pos(S1, S2);
end;

{TRepToDateExpression}

function TRepToDateExpression.DefFunction(var AName: String): Integer;
begin
  AName := 'TODATE';
  Result := 1;
end;

function TRepToDateExpression.GetAsVariant: Variant;
begin
  Result := StrToDate(Values[0]);
end;

{TRepToNumberExpression}

function TRepToNumberExpression.DefFunction(var AName: String): Integer;
begin
  AName := 'TONUMBER';
  Result := 1;
end;

function TRepToNumberExpression.GetAsVariant: Variant;
begin
  Result := StrToFloat(Values[0]);
end;

{TRepToStringExpression}

function TRepToStringExpression.DefFunction(var AName: String): Integer;
begin
  AName := 'TOCHAR';
  Result := 1;
end;

function TRepToStringExpression.GetAsVariant: Variant;
var
  S: String;
begin
  S := Values[0];
  Result := S;
end;

{TRepPageHeightExpression}

procedure TRepPageHeightExpression.Prepare;
begin
  // do nothing
end;

function TRepPageHeightExpression.GetAsVariant: Variant;
begin
  Result := RepEngine.PageHeight;
end;

{TRepExtraHeightExpression}

procedure TRepExtraHeightExpression.Prepare;
begin
  FPInt := RepEngine.GetSysAddress(SYS_ADDR_EXTRA_HEIGHT);
end;

function TRepExtraHeightExpression.GetAsVariant: Variant;
begin
  Result := -(FPInt)^;
end;

{TRepExtraHeightUsedExpression}

procedure TRepExtraHeightUsedExpression.Prepare;
begin
  FPInt := RepEngine.GetSysAddress(SYS_ADDR_EXTRA_USED);
end;

function TRepExtraHeightUsedExpression.GetAsVariant: Variant;
begin
  Result := -(FPInt^);
end;

{TRepCurPageHeightExpression}

procedure TRepCurPageHeightExpression.Prepare;
begin
  FPInt := RepEngine.GetSysAddress(SYS_ADDR_CURPAGEHEIGHT);
end;

function TRepCurPageHeightExpression.GetAsVariant: Variant;
begin
  Result := FPInt^;
end;

{TRepCtrlHeightExpr}

procedure TRepCtrlHeightExpr.Prepare;
begin
  if SelfPrepared then exit;
  FControl := RepEngine.ControlByToken[FExpr];
  SelfPrepared := True;
end;

function TRepCtrlHeightExpr.GetAsVariant: Variant;
begin
  Result := FControl.Height;
end;

{TRepRoundExpression}

function TRepRoundExpression.DefFunction(Var AName: String): integer;
begin
  AName := 'ROUND';
  Result := 2;
end;

function TRepRoundExpression.GetAsVariant: Variant;
var
  Exponent: Integer;
  Val: Extended;
begin
  Val := Values[0];
  Exponent := Values[1];
  Result := Round(Val*IntPower(10,Exponent))/(IntPower(10,Exponent));
end;

{TRepTruncExpression}

function TRepTruncExpression.DefFunction(Var AName: String): integer;
begin
  AName := 'TRUNC';
  Result := 2;
end;

function TRepTruncExpression.GetAsVariant: Variant;
var
  Exponent: Integer;
  Val: Extended;
begin
  Val := Values[0];
  Exponent := Values[1];
  Result := Trunc(Val*IntPower(10,Exponent))/(IntPower(10,Exponent));
end;

{TRepLengthExpression}

function TRepLengthExpression.DefFunction(Var AName: String): integer;
begin
  AName := 'LENGTH';
  Result := 1;
end;

function TRepLengthExpression.GetAsVariant: Variant;
begin
  Result := Length(Values[0]);
end;

{TRepLTrimExpression}

function TRepLTrimExpression.DefFunction(Var AName: String): integer;
begin
  AName := 'LTRIM';
  Result := 1;
end;

function TRepLTrimExpression.GetAsVariant: Variant;
begin
  Result := TrimLeft(Values[0]);
end;

{TRepRTrimExpression}

function TRepRTrimExpression.DefFunction(Var AName: String): integer;
begin
  AName := 'RTRIM';
  Result := 1;
end;

function TRepRTrimExpression.GetAsVariant: Variant;
begin
  Result := TrimRight(Values[0]);
end;

{TRepTrimExpression}

function TRepTrimExpression.DefFunction(Var AName: String): integer;
begin
  AName := 'TRIM';
  Result := 1;
end;

function TRepTrimExpression.GetAsVariant: Variant;
begin
  Result := Trim(Values[0]);
end;

{TRepLocalDate}

function TRepLocalDate.GetAsVariant: Variant;
begin
  Result := Now;
end;

procedure TRepLocalDate.Prepare;
begin
  {}
end;

{TRepLocalTime}

function TRepLocalTime.GetAsVariant: Variant;
begin
  Result := Time;
end;

procedure TRepLocalTime.Prepare;
begin
  {}
end;

{TRepNull}

function TRepNull.GetAsVariant: Variant;
begin
  Result := Null;
end;

procedure TRepNull.Prepare;
begin
  {}
end;

initialization
  AddPlugIn('CHECK', [akRight], TRepCheckExpression, False);
  AddPlugIn('SUM', [akRight], TRepAccumulativeSum, False);
  AddPlugIn('MAX', [akRight], TRepAccumulativeMax, False);
  AddPlugIn('MIN', [akRight], TRepAccumulativeMin, False);
  AddPlugIn('AVG', [akRight], TRepAccumulativeAvg, False);
  AddPlugIn('+', [akLeft, akRight], TRepBinPlusExpr, True);
  AddPlugIn('||', [akLeft, akRight], TRepBinPlusExpr, True);
  AddPlugIn('-', [akLeft, akRight], TRepBinMinusExpr, True);
  AddPlugIn('*', [akLeft, akRight], TRepBinMulExpr, True);
  AddPlugIn('/', [akLeft, akRight], TRepBinDivExpr, True);
  AddPlugIn('AND', [akLeft, akRight], TRepBinAndExpr, False);
  AddPlugIn('OR', [akLeft, akRight], TRepBinOrExpr, False);
  AddPlugIn('XOR', [akLeft, akRight], TRepBinXorExpr, False);
  AddPlugIn('-', [akRight], TRepUnaryMinusExpr, True);
  AddPlugIn('NOT', [akRight], TRepNotExpr, False);
  AddPlugIn('->', [akLeft, akRight], TRepFieldExpression, True);
  AddPlugIn('>', [akLeft, akRight], TRepMoreExpr, True);
  AddPlugIn('>=', [akLeft, akRight], TRepMoreOrEquelExpr, True);
  AddPlugIn('<', [akLeft, akRight], TRepLessExpr, True);
  AddPlugIn('<=', [akLeft, akRight], TRepLessOrEquelExpr, True);
  AddPlugIn('<>', [akLeft, akRight], TRepNotEquelExpr, True);
  AddPlugIn('=', [akLeft, akRight], TRepEquelExpr, True);
  AddPlugIn('=>', [akLeft, akRight], TSubRepExpression, True);
  AddPlugIn('OLD', [akRight], TRepPrevFiledValExpr, False);
  AddPlugIn('PAGESTART', [], TPageStartExpression, False);
  AddPlugIn('PAGEFINISH', [], TPageEndExpression, False);
  AddPlugIn('REPORTSTART', [], TRepStartExpression, False);
  AddPlugIn('REPORTFINISH', [], TRepEndExpression, False);
  AddPlugIn('PAGENUM', [], TPageCountExpression, False);
  AddPlugIn(',', [akLeft, akRight], TRepUnaryExpression, True);
  AddPlugIn('IF', [akRight, akSet], TRepCaseExpression, False);
  AddPlugIn('NVL', [akRight, akSet], TRepNvlExpression, False);
  AddPlugIn('SINGLEROWED', [akRight], TRepQueryOpenedExpr, False);
  AddPlugIn('ROWSFETCHED', [akRight], TRepRowsFetchedCountExpr, False);
  AddPlugIn('EOF', [akRight], TRepEOFExpression, False);
  AddPlugIn('EMPTY', [akRight], TEmptyQueryExpression, False);
  AddPlugIn('ERRORSUBST', [akRight, akSet], TRepErrorSubstExpr, False);
  AddPlugIn('DEBUG', [akRight, akSet], TRepDebugExpr, False);
  AddPlugIn('LEFTSTR', [akRight, akSet], TRepLeftStrExpr, False);
  AddPlugIn('RIGHTSTR', [akRight, akSet], TRepRightStrExpr, False);
  AddPlugIn('SUBSTR', [akRight, akSet], TRepSubStrExpr, False);
  AddPlugIn(':=', [akLeft, akRight], TRepAssignExpr, True);
  AddPlugIn('REPLICATE', [akRight, akSet], TRepReplicateExpr, False);
  AddPlugIn('FORMAT', [akRight, akSet], TRepFormatExpr, False);
  AddPlugIn('CONTROLPROCESSED', [akRight], TRepCtrlProcessedExpr, False);
  AddPlugIn('QUERYPROCESSED', [akRight], TRepQueryProcessedExpr, False);
  AddPlugIn('SUBREPORTPROCESSED', [akRight], TSubRepProcessedExpr, False);
  AddPlugIn('CHR', [akRight, akSet], TRepChrExpression, False);
  AddPlugIn('DECODE', [akRight, akSet], TRepDecodeExpression, False);
  AddPlugIn('POS', [akRight, akSet], TRepPosExpression, False);
  AddPlugIn('TODATE', [akRight, akSet], TRepToDateExpression, False);
  AddPlugIn('TONUMBER', [akRight, akSet], TRepToNumberExpression, False);
  AddPlugIn('TOCHAR', [akRight, akSet], TRepToStringExpression, False);
  AddPlugIn('PAGEHEIGHT', [], TRepPageHeightExpression, False);
  AddPlugIn('CURPAGEHEIGHT', [], TRepCurPageHeightExpression, False);
  AddPlugIn('EXTRAHEIGHT', [], TRepExtraHeightExpression, False);
  AddPlugIn('EXTRAUSED', [], TRepExtraHeightUsedExpression, False);
  AddPlugIn('CONTROLHEIGHT', [akRight], TRepCtrlHeightExpr, False);
  AddPlugIn('ROUND', [akRight, akSet], TRepRoundExpression, False);
  AddPlugIn('TRUNC', [akRight, akSet], TRepTruncExpression, False);
  AddPlugIn('LENGTH', [akRight, akSet], TRepLengthExpression, False);
  AddPlugIn('LTRIM', [akRight, akSet], TRepLTrimExpression, False);
  AddPlugIn('RTRIM', [akRight, akSet], TRepRTrimExpression, False);
  AddPlugIn('TRIM', [akRight, akSet], TRepTrimExpression, False);
  AddPlugIn('LOCALDATE', [], TRepLocalDate, False);
  AddPlugIn('LOCALTIME', [], TRepLocalTime, False);
  AddPlugIn('NULL', [], TRepNull, False);
end.
