{**
*
*         Script Reporter Component library
*         Kernel module
*          Vlad Ilushchenko 1998-99
*
*
**}

unit RepKrnl;

interface

uses Classes, ObjList, Oracle, SysUtils, Graphics, DBTables, DB;

type
  TRepExpression   = class;
  TCustomRepEngine = class;
  TRepControl      = class;

  TPageState         = (psStarting,
                        psStarted,
                        psFinishing,
                        psSafeFinishing,
                        psFinished);
  TReportState = (rsParsed, rsPrepared, rsComleted);

  TRepQueryParamType = (ptParam, ptMacro);
  TTerminateType     = (ttNormal, ttAbort, ttError);

  TTerminateEvent = procedure (Sender: TCustomRepEngine;
    TerminateType: TTerminateType) of object;

  EReportTerminated = class(Exception)
  private
    FErrorCode: Integer;
  public
    constructor Create(const Msg: String; Code: Integer);
    property ErrorCode: Integer read FErrorCode;
  end;

  TRepPosition = record
    Line: Integer;
    Col : Integer;
  end;

  TRepToken = record
    Pos: TRepPosition;
    Text: string;
    Addr: TRepExpression;
    AddrStr: String;
  end;

  TRepVersion = record
    Major:   Integer;
    Minor:   Integer;
    Release: Integer;
    Comment: String;
  end;

  TRepObject = class
  private
    FRepEngine: TCustomRepEngine;
    FName: String;
    FProcessed: Boolean;
    FResetClause: TRepToken;
    FWhenClause:  TRepToken;
    FSelfPrepared: Boolean;
    FPosition: TRepPosition;
  protected
    procedure Turn; virtual;
    procedure Prepare; virtual;
    function DoTurn: Boolean; virtual;
    procedure BeginTurn; virtual;
    procedure EndTurn; virtual;
  public
    constructor CreateOriginal(ARepEngine: TCustomRepEngine;
      AWhenClause, AResetClause: TRepToken; AName: String;
      Pos: TRepPosition);

    function MakeTurn: Boolean;
    procedure FinishTurn;
    property Name:        String         read FName;
    property Processed:   Boolean        read FProcessed;
    property RepEngine:   TCustomRepEngine  read FRepEngine;
    property WhenClause:  TRepToken read FWhenClause write FWhenClause;
    property ResetClause: TRepToken read FResetClause write FResetClause;
    property SelfPrepared: Boolean read FSelfPrepared write FSelfPrepared;
    property ScriptPos: TRepPosition read FPosition;
  end;

  TRepExpressionClass = class of TRepExpression;

  TRepExpression = class(TRepObject)
  private
    FTurnable: Boolean;
    FResetBefore: Boolean;
    FResetAfter: Boolean;
  protected
    function GetAsVariant: Variant; virtual;
    procedure SetAsVariant(Value: Variant); virtual;
    procedure Turn; override;
    function DoTurn: Boolean; override;
    procedure SetTurnable(Value: Boolean); virtual;
    function IntGetAsVariant: Variant;
    function IntGetOldValue: Variant;
    function GetOldValue: Variant; virtual;
    procedure Prepare; override;
    function Reset: Boolean; virtual;
  public
    constructor CreateOriginal(RepEngine: TCustomRepEngine;
      WhenClause, ResetClause: TRepToken; AName: String; Pos: TRepPosition);
    procedure Bind;
    property Value: Variant read IntGetAsVariant write SetAsVariant;
    property Turnable: Boolean read FTurnable write SetTurnable;
    property ResetBefore: Boolean read FResetBefore write FResetBefore;
    property ResetAfter: Boolean read FResetAfter write FResetAfter;
    property OldValue: Variant read IntGetOldValue;
  end;

  TRepQueryParam = class
  private
    FName: String;
    FExpr: TRepExpression;
    FType: TRepQueryParamType;
  public
    constructor CreateNew(Name: String; Expr: TRepExpression;
      AType: TRepQueryParamType);
  end;

  TRepQuery = class(TRepObject)
  private
    FQuery        : TOracleQuery;
    FCursorQuery  : TOracleQuery;
    FBDEQuery     : TDataSet;
    FParams       : TObjectList;
    FSQL          : TStringList;
    FNullFields   : Boolean;
    FExecuted     : Boolean;
    FRowsFetched  : Integer;
    FTurned       : Boolean;
    FModifyParams : Boolean;
    FUnderBDE     : Boolean;
    FAlias        : String;
    FRefName      : String;
    FHasReference : Boolean;
    FHasCursor    : Boolean;
    FHasMacros    : Boolean;
    FCursorName   : String;
    FSQLPosition  : TRepPosition;
    procedure ApplyParams;
    procedure ApplyQueryParamsToExpr;
    procedure SetRefName(Value: String);
    function GetField(FieldName: String): Variant;
    function GetParams(Index: Integer): TRepQueryParam;
    function GetParamCount: Integer;
    function GetEOF: Boolean;
    function GetFetchBy: Integer;
    procedure SetFetchBy(Value: Integer);
    function GetErrorPosition: TRepPosition;
  protected
    function GetSQL: TStrings;
    function DoTurn: Boolean; override;
    procedure SetSQL(Value: TStrings);
    procedure Open; virtual;
    procedure Close; virtual;
    procedure Prepare; override;
    procedure EnumVariables(VarName: String; Expr: TRepExpression;
      ParType: TRepQueryParamType);
    procedure Turn; override;
    procedure DescribeMacros(StringList: TStrings);
    procedure BindMacros(StringList: TStrings);

    property Params[Index: Integer]: TRepQueryParam read GetParams;
    property ParamCount: Integer read GetParamCount;
    property Query: TOracleQuery read FQuery;
    property BDEQuery: TDataSet read FBDEQuery;
  public
    constructor CreateOriginal(RepEngine: TCustomRepEngine;
      WhenClause, ResetClause: TRepToken; AName: String; CursorName: String);
    destructor Destroy; override;

    procedure BeginTurn; override;
    property Fields[Name: String]: Variant read GetField;
    property SQL: TStrings read GetSQL write SetSQL;
    property NullFields: Boolean read FNullFields write FNullFields;
    property Executed: Boolean read FExecuted;
    property RowsFetched: Integer read FRowsFetched;
    property ModifyParams: Boolean read FModifyParams write FModifyParams;
    property EOF: Boolean read GetEOF;
    property Alias: String read FAlias write FAlias;
    property UnderBDE: Boolean read FUnderBDE write FUnderBDE;
    property RefName: String read FRefName write SetRefName;
    property HasReference: Boolean read FHasReference write FHasReference;
    property FetchBy: Integer read GetFetchBy write SetFetchBy;
    property SQLPosition: TRepPosition read FSQLPosition write FSQLPosition;
    property ErrorPosition: TRepPosition read GetErrorPosition;
  end;

  TRepControl = class(TRepObject)
  private
    FOrder          : Integer;
    FAlwaysOnPage   : Boolean;
    FFinishesPage   : Boolean;
    FDisableFooter  : Boolean;
    FProcessedCount : Integer;
    procedure SetOrder(Value: Integer);
  protected
    function DoTurn: Boolean; override;
    function GetHeight: Integer; virtual; abstract;
    procedure Print(AText: String; AFont: TFont);
    procedure Prepare; override;
  public
    constructor CreateOriginal(RepEngine: TCustomRepEngine;
      WhenClause: TRepToken; AName: String);
    destructor Destroy; override;

    property Height: Integer read GetHeight;
    property Order: Integer read FOrder write SetOrder;
    property AlwaysOnPage: Boolean read FAlwaysOnPage write FAlwaysOnPage;
    property FinishesPage: Boolean read FFinishesPage write FFinishesPage;
    property DisableFooter: Boolean read FDisableFooter write FDisableFooter;
    property ProcessedCount: Integer read FProcessedCount write FProcessedCount;
  end;

  TRepParams = class;

  TRepParam = class(TPersistent)
  private
    FParamList: TRepParams;
    FValue: Variant;
    FParamName: String;
    FInputMask: String;
    FPosition: TRepPosition;
    procedure SetValue(AValue: Variant);
    function GetDataSize: Integer;
    procedure GetData(Buffer: Pointer);
    procedure SetData(Buffer: Pointer; DataType: Word);
  public
    constructor Create(AParamList: TRepParams; AName: String; Pos: TRepPosition);
    destructor Destroy; override;

    procedure Assign(Source: TPersistent); override;
    property Value: Variant read FValue write SetValue;
    property ParamName: String read FParamName;
    property InputMask: String read FInputMask write FInputMask;
  end;

  TRepParams = class(TPersistent)
  private
    FList: TObjectList;
    FRepEngine: TCustomRepEngine;
    function GetParamCount: Integer;
    function GetParamValue(const ParamName: String): Variant;
    function GetParam(Index: Integer): TRepParam;
    procedure SetParamValue(const ParamName: String; Value: Variant);
    procedure ReadBinaryData(Stream: TStream);
    procedure WriteBinaryData(Stream: TStream);
  protected
    procedure DefineProperties(Filer: TFiler); override;
  public
    constructor Create(ARepEngine: TCustomRepEngine); virtual;
    destructor Destroy; override;
    procedure Assign(Source: TPersistent); override;
    procedure Clear;
    procedure Bind;
    procedure AddParam(Value: TRepParam);
    procedure RemoveParam(Value: TRepParam);
    function CreateParam(AName: String; Pos: TRepPosition): TRepParam;
    function ParamByName(const AParamName: String): TRepParam;
    property ParamValues[const ParamName: String]: Variant read
      GetParamValue write SetParamValue; default;
    property Param[Index: Integer]: TRepParam read GetParam;
    property ParamCount: Integer read GetParamCount;
  end;

  TReportThread = class(TThread)
  private
    FRepEngine: TCustomRepEngine;
    FSemaphore: Integer;
  protected
    procedure SignalForPage;
    procedure WaitForPage;
    procedure DoOnNextPage;
    procedure DoOnReportStart;
    procedure SyncDoBeforeSubRepParsed;
  protected
    constructor Create(RepEngine: TCustomRepEngine);
    procedure Execute; override;
  public
    destructor Destroy; override;
  end;

  TCustomRepEngine = class(TComponent)
  private
    // variuos lists
    FExprList: TObjectList;
    FTurnableExprList: TObjectList;
    FQueryList: TObjectList;
    FControlList: TObjectList;
    FChildRepList: TObjectList;
    FParams: TRepParams;
    FErrorList: TStrings;

    FParentReport: TCustomRepEngine;
    FOracleSession: TOracleSession;
    // Top value for semaphore
    FMaxPages: Longint;

    // Thread object for handling separate execution
    FReportThread: TReportThread;
    FMultiThreaded: Boolean;

    // WHEN cluse, used when this object works as subreport
    FWhenClause: TRepToken;
    FNoWait: Boolean;
    FRepProcessed: Boolean;
    // Page handling
    FPageHeight: Integer;
    FExtraHeight: Integer;
    FExtraHeightUsed: Integer;
    FCurPageHeight: Integer;
    FPageState: TPageState;
    FPageCount: Integer;

    // Report handling
    FRepStarted: Boolean;
    FRepFinished: Boolean;
    FQueryFinished: Boolean;

    // some variables for linking scripts
    FPrepared: Boolean;
    FParsed: Boolean;
    FPreparedObj: TRepObject;
    FAddParamToList: Boolean;
    FListFreshed: Boolean;
    // Events
    FOnTerminate: TTerminateEvent;
    FOnNextPage: TNotifyEvent;
    FOnStart: TNotifyEvent;
    // Misc.
    FChildRepIndex: Integer;
    function GetCommonRepExpr(Index: Integer): TRepExpression;
    function GetTurnableRepExpr(Index: Integer): TRepExpression;
    function GetExprByToken(AToken: TRepToken): TRepExpression;
    function GetRepExprName(AName: String): TRepExpression;
    function GetCommonRepExprCount: Integer;
    function GetTurnableRepExprCount: Integer;
    function GetRepQuery(Index: Integer): TRepQuery;
    function GetQueryByToken(AToken: TRepToken): TRepQuery;
    function GetRepQueryName(AName: String): TRepQuery;
    function GetRepQueryCount: Integer;
    function GetRepControl(Index: Integer): TRepControl;
    function GetControlByToken(AToken: TRepToken): TRepControl;
    function GetRepControlName(AName: String): TRepControl;
    function GetRepControlCount: Integer;
    function GetChildRep(Index: Integer): TCustomRepEngine;
    function GetChildRepName(AToken: TRepToken): TCustomRepEngine;
    function GetChildRepCount: Integer;
    function GetTerminated: Boolean;
    function GetHasErrors: Boolean;
    procedure SetParamsList(Value: TRepParams);
    procedure PrepareSubreports;
    procedure ParseSubreports;
    procedure InternalRun;
    procedure ReportTerminated(Sender: TObject);
    procedure ClearErrorList;
  protected
    procedure RefreshCtrlList(ARepControl: TRepControl);
    procedure PrepareList(AList: TObjectList);
    procedure EvalExpressions;
    procedure ResetExpressions;
    function ProcessQueries: Boolean;
    function PrintControls: Boolean;
    procedure RunSubreports;
    procedure ProcessEndTurn;
    procedure WaitForSubreports;

    procedure Print(AText: String; AFont: TFont); virtual; abstract;
    procedure Skip(Units: Integer; ToNewPage: Boolean); virtual; abstract;

    procedure Prepare; virtual;
    procedure UnPrepare; virtual;
    procedure RefreshList;
    procedure Parse; virtual; abstract;
    procedure ResetReport;
    procedure SetPrepared(Value: Boolean); virtual;
    function MakeTurn: Boolean;
    procedure Run;
    procedure OpenErrors; virtual; abstract;
    procedure OpenCompiler; virtual; abstract;
    procedure DoBeforeSubRepParsed(Index: Integer); virtual; abstract;

    property CommonRepExpr[Index: Integer]: TRepExpression read GetCommonRepExpr;
    property CommonRepExprCount: Integer read GetCommonRepExprCount;
    property TurnableRepExpr[Index: Integer]: TRepExpression read GetTurnableRepExpr;
    property TurnableRepExprCount: Integer read GetTurnableRepExprCount;
    property ChildRep[Index: Integer]: TCustomRepEngine read GetChildRep;
    property AddParamToList: Boolean read FAddParamToList;
    property WhenClause: TRepToken read FWhenClause write FWhenClause;
  public
    constructor Create(AOwner: TComponent); override;
    constructor CreateSubreport(AOwner: TComponent; AParentReport: TCustomRepEngine);
      virtual;
    destructor Destroy; override;

    function RepVersion: TRepVersion;

    procedure PrepareReport(ParseIt: Boolean);
    procedure GetParams;
    procedure ReportError(ErrorCode: Word;
                          ErrorObjectName, ExtraData: String;
                          RepObject: TRepObject; RaiseException: Boolean);

    function AddExpression(Expr: TRepExpression): TRepExpression;
    function AddQuery(AQuery: TRepQuery): TRepQuery;
    function AddControl(AControl: TRepControl): TRepControl;
    function AddSubreport(ASubreport: TCustomRepEngine): TCustomRepEngine;

    procedure WaitForTerminate;
    procedure WaitForNextPage;
    procedure Terminate;

    function GetSysAddress(AddressType: Word): Pointer;

    property PreparedObject: TRepObject read FPreparedObj write FPReparedObj;
    property PageHeight: Integer read FPageHeight write FPageHeight;
    property ExprByToken[AToken: TRepToken]: TRepExpression read GetExprByToken;
    property RepExprName[AName: String]: TRepExpression read GetRepExprName;
    property RepQuery[Index: Integer]: TRepQuery read GetRepQuery;
    property QueryByToken[AToken: TRepToken]: TRepQuery read GetQueryByToken;
    property RepQueryName[AName: String]: TRepQuery read GetRepQueryName;
    property RepQueryCount: Integer read GetRepQueryCount;
    property RepControl[Index: Integer]: TRepControl read GetRepControl;
    property ControlByToken[AToken: TRepToken]: TRepControl read GetControlByToken;
    property RepControlName[AName: String]: TRepControl read GetRepControlName;
    property RepControlCount: Integer read GetRepControlCount;
    property ChildRepName[AToken: TRepToken]: TCustomRepEngine read GetChildRepName;
    property ChildRepCount: Integer read GetChildRepCount;
    property OracleSession: TOracleSession read FOracleSession
      write FOracleSession;
    property RepStarted: Boolean read FRepStarted;
    property RepFinished: Boolean read FRepFinished;
    property PageCount: Integer read FPageCount;
    property Params: TRepParams read FParams write SetParamsList;
    property Prepared: Boolean read FPrepared write SetPrepared stored False;
    property RepProcessed: Boolean read FRepProcessed write FRepProcessed;
    property ParentReport: TCustomRepEngine read FParentReport;
    property MultiThreaded: Boolean read FMultiThreaded write FMultiThreaded;
    property MaxPages: Longint read FMaxPages write FMaxPages;
    property Terminated: Boolean read GetTerminated;
    property NoWait: Boolean read FNoWait write FNoWait;
    property HasErrors: Boolean read GetHasErrors;
    property ErrorList: TStrings read FErrorList;
    property OnTerminate: TTerminateEvent read FOnTerminate write FOnTerminate;
    property OnNextPage: TNotifyEvent read FOnNextPage write FOnNextPage;
    property OnStart: TNotifyEvent read FOnStart write FOnStart;
  end;

  {Here are a special kinds of repControl's}

  TPageEndExpression = class(TRepExpression)
  protected
    function GetAsVariant: Variant; override;
  public
    constructor Create(RepEngine: TCustomRepEngine; Pos: TRepPosition);
  end;

  TPageStartExpression = class(TRepExpression)
  protected
    function GetAsVariant: Variant; override;
  public
    constructor Create(RepEngine: TCustomRepEngine; Pos: TRepPosition);
  end;

  TRepEndExpression = class(TRepExpression)
  protected
    function GetAsVariant: Variant; override;
  public
    constructor Create(RepEngine: TCustomRepEngine; Pos: TRepPosition);
  end;

  TRepStartExpression = class(TRepExpression)
  protected
    function GetAsVariant: Variant; override;
  public
    constructor Create(RepEngine: TCustomRepEngine; Pos: TRepPosition);
  end;

  TPageCountExpression = class(TRepExpression)
  protected
    function GetAsVariant: Variant; override;
  public
    constructor Create(RepEngine: TCustomRepEngine; Pos: TRepPosition);
  end;

  EParamNotFound = class(Exception) end;
  EParamExist = class(Exception) end;

const
  SYS_ADDR_CURPAGEHEIGHT = 0;
  SYS_ADDR_EXTRA_HEIGHT  = 1;
  SYS_ADDR_EXTRA_USED    = 2;

function VeryGlobalFindComponent(Name: String): TComponent;
function RepPosition(Line, Col: Integer): TRepPosition;
function EmptyRepToken: TRepToken;
function Str2RepToken(Str: String): TRepToken;
function ToRepToken(Text: String; Addr: TRepExpression; AddrStr: String;
                    Pos: TRepPosition): TRepToken;
function NameToRepToken(AName: String): TRepToken;

implementation

{$R errmsgs.res}

uses RepExpr, Forms, Dialogs, Windows, RepDebugger, OracleCI, BDE, DBConsts;

function RepPosition(Line, Col: Integer): TRepPosition;
begin
  Result.Line := Line;
  Result.Col  := Col;
end;

function CompareRepCtrlsOrder(Item1, Item2: Pointer): Integer; assembler;
asm
  mov eax, [eax].TRepControl.FOrder
  mov edx, [edx].TRepControl.FOrder
  sub eax, edx
end;

procedure DescribeParams(SQL: TStrings; OutList: TStrings);
var
  i: Integer;
  s, s1: String;
  ParamName: String;
  l: Integer;
begin
  OutList.Clear;
  for i := 0 to SQL.Count-1 do
  begin
    s := SQL.Strings[i];

    // cut string constants

    while True do
    begin
      l := Pos('''', s);
      if l = 0 then break;

      s1 := Copy(s, 1, l-1);
      s := Copy(s, l + 1, Length(s));

      l := Pos('''', s);
      if l = 0 then
        s := ''
      else
        s := Copy(s, l + 1, Length(s));
      s := s1 + s;
    end;

    // cut "havy" comments

    while True do
    begin
      l := Pos('/*', s);
      if l = 0 then break;

      s1 := Copy(s, 1, l-1);
      s := Copy(s, l + 1, Length(s));

      l := Pos('*/', s);

      if l = 0 then
        s := ''
      else
        s := Copy(s, l + 1, Length(s));
      s := s1 + s;
    end;

    // extract pure placeholder

    while Pos(':', s) > 0 do
    begin
      ParamName := '';
      s := copy(s, Pos(':', s) + 1, length(s));
      if length(s) = 0 then break;

      while ((s[1]>='a')and(s[1]<='z'))or
            ((s[1]>='A')and(s[1]<='Z'))or
            ((s[1]>='0')and(s[1]<='9'))or
            (s[1]='_') do
      begin
        ParamName := ParamName + s[1];
        s := copy(s, 2, length(s));
        if length(s) = 0 then break;
      end;

      // just a semicolon
      if ParamName = '' then continue;

      ParamName := AnsiLowerCase(ParamName);
      if OutList.IndexOf(ParamName) = -1 then OutList.Add(ParamName);
    end;
  end;
end;

{EReportTerminated}

constructor EReportTerminated.Create(const Msg: String; Code: Integer);
begin
  inherited Create(Msg);
  FErrorCode := Code;
end;

{TRepObject}

constructor TRepObject.CreateOriginal(ARepEngine: TCustomRepEngine;
  AWhenClause, AResetClause: TRepToken; AName: String;
  Pos: TRepPosition);
begin
  inherited Create;
  FRepEngine   := ARepEngine;
  FProcessed   := False;
  FWhenClause  := AWhenClause;
  FResetClause := AResetClause;
  FPosition    := Pos;
  FName := AName;
end;

procedure TRepObject.BeginTurn;
begin
end;

procedure TRepObject.EndTurn;
begin
  FProcessed := False;
end;

function TRepObject.MakeTurn: Boolean;
begin
  Result := False;
  try
    Result := DoTurn;
  except
    on E: Exception do
      RepEngine.ReportError(70, Self.Name, E.Message, Self, True);
  end;
end;

function TRepObject.DoTurn: Boolean;
var
  WhenResult: Boolean;
begin
  Result := not Processed;

  if Processed then exit;

  if WhenClause.Addr <> nil then
    WhenResult := WhenClause.Addr.Value
  else
    WhenResult := False;

  if not WhenResult and
     not ((WhenClause.Addr = nil) and (ResetClause.Addr = nil)
          and RepEngine.RepFinished )then
    Turn;

  FProcessed := True;
end;

procedure TRepObject.Turn;
begin
  FProcessed := True;
end;

procedure TRepObject.Prepare;
begin
  if SelfPrepared then exit;

  if (WhenClause.Text <> EmptyStr) and (WhenClause.Addr = nil) then
  begin
    RepEngine.FPreparedObj := Self;
    FWhenClause.Addr := RepEngine.ExprByToken[FWhenClause];
  end;

  if (FResetClause.Text <> EmptyStr) and (FResetClause.Addr = nil) then
  begin
    RepEngine.FPreparedObj := Self;
    FResetClause.Addr := RepEngine.ExprByToken[FResetClause]
  end;
  SelfPrepared := True;
end;

procedure TRepObject.FinishTurn;
begin
  EndTurn;
end;

{TRepExpression}

constructor TRepExpression.CreateOriginal(RepEngine: TCustomRepEngine;
     WhenClause, ResetClause: TRepToken; AName: String; Pos: TRepPosition);
begin
  inherited CreateOriginal(RepEngine,
                           WhenClause,
                           ResetClause,
                           AName, Pos);
  FTurnable := False;
  FResetBefore := False;
  FResetAfter := False;
end;

procedure TRepExpression.SetTurnable(Value: Boolean);
var
  SourceList: TObjectList;
  DestList: TObjectList;
  i: Integer;
begin
  if FTurnable = Value then exit;
  if RepEngine = nil then exit;

  with RepEngine do
    if FTurnable then
    begin
      SourceList := FTurnableExprList;
      DestList   := FExprList;
    end
    else
    begin
      SourceList := FExprList;
      DestList := FTurnableExprList;
    end;

  FTurnable := Value;

  i := SourceList.IndexOf(Self);

  if i = -1 then exit;

  SourceList.SafeDelete(i);
  DestList.Add(Self);
  RepEngine.RefreshList;

end;

function TRepExpression.Reset: Boolean;
begin
  Result := False;
  if (WhenClause.Addr <> nil) and (ResetClause.Addr <> nil) then
    if WhenClause.Addr.Value = True then
    begin
      Self.Value := ResetClause.Addr.Value;
      Result := True;
    end;
end;

procedure TRepExpression.Prepare;
begin
  if SelfPrepared then exit;

  inherited Prepare;
  if ResetClause.Addr <> nil then
    Self.Value := ResetClause.Addr.Value;
  SelfPrepared := True;
end;

procedure TRepExpression.SetAsVariant(Value: Variant);
begin
  RepEngine.ReportError(1, Self.ClassName, Self.Name, Self, True);
end;

function TRepExpression.GetAsVariant: Variant;
begin
  RepEngine.ReportError(2, Self.ClassName, Self.Name, Self, True);
end;

function TRepExpression.DoTurn: Boolean;
var
  WhenResult: Boolean;
begin
  Result := not Processed;

  if Processed then exit;

  if WhenClause.Addr <> nil then
    WhenResult := WhenClause.Addr.Value
  else
    WhenResult := False;

  if (not WhenResult) or ResetAfter {and
     not ((WhenClause = nil) and (ResetClause = nil) and RepEngine.RepFinished)}
     then
    Turn
  else if WhenResult and FResetBefore then
    if Reset then
      Turn;

  FProcessed := True;
end;

procedure TRepExpression.Turn;
begin
  inherited Turn;
end;

function TRepExpression.GetOldValue: Variant;
begin
  Result := GetAsVariant;
end;

function TRepExpression.IntGetOldValue: Variant;
begin
  if Turnable and (not RepEngine.RepFinished) then MakeTurn;
  try
    Result := GetOldValue;
  except
    on E: Exception do
      RepEngine.ReportError(4, Self.Name, E.Message, Self, True);
  end;
end;

function TRepExpression.IntGetAsVariant: Variant;
begin
  try
    if Turnable and (not RepEngine.RepFinished) then MakeTurn;
    Result := GetAsVariant;
  except
    on E: Exception do
      RepEngine.ReportError(5, Self.Name, E.Message, Self, True);
  end;
end;

procedure TRepExpression.Bind;
begin
  Prepare;
end;

{TRepQueryParam}

constructor TRepQueryParam.CreateNew(Name: String; Expr: TRepExpression;
  AType: TRepQueryParamType);
begin
  inherited Create;
  FName := Name;
  FExpr := Expr;
  FType := AType;
end;

{TRepControl}

constructor TRepControl.CreateOriginal(RepEngine: TCustomRepEngine;
                                       WhenClause: TRepToken; AName: String);
begin
   inherited CreateOriginal(RepEngine,
                            WhenClause,
                            EmptyRepToken,
                            AName,
                            RepPosition(-1,-1));
   FAlwaysOnPage := False;
   FProcessedCount := 0;
end;

destructor TRepControl.Destroy;
begin
  inherited Destroy;
end;

procedure TRepControl.SetOrder(Value: Integer);
begin
  if FOrder = Value then exit;
  FOrder := Value;
  FRepEngine.RefreshCtrlList(Self);
end;

procedure TRepControl.Print(AText: String; AFont: TFont);
begin
  RepEngine.Print(AText, AFont);
end;

procedure TRepControl.Prepare;
begin
  if SelfPrepared then exit;

  inherited Prepare;
  if AlwaysOnPage then Dec(RepEngine.FExtraHeight, Height);
  SelfPrepared := True;
end;

function TRepControl.DoTurn: Boolean;
var
  WhenResult: Boolean;
begin
  Result := FAlwaysOnPage or (not Processed);

  if (not FAlwaysOnPage) and Processed then exit;

  if WhenClause.Addr <> nil then
    WhenResult := WhenClause.Addr.Value
  else
    WhenResult := True;

  if WhenResult and
     not ((WhenClause.Addr = nil) and (ResetClause.Addr = nil) and
           RepEngine.RepFinished )then
  begin
    Turn;
    Inc(FProcessedCount);
  end;

  if not FAlwaysOnPage then
    FProcessed := True;
end;

{TRepQuery}

constructor TRepQuery.CreateOriginal(RepEngine: TCustomRepEngine;
  WhenClause, ResetClause: TRepToken; AName: String; CursorName: String);
begin
  inherited CreateOriginal(RepEngine, WhenClause, ResetClause, AName, RepPosition(-1,-1));
  FQuery        := TOracleQuery.Create(RepEngine);
  FCursorQuery  := TOracleQuery.Create(RepEngine);
  FBDEQuery     := TQuery.Create(RepEngine);
  FParams       := TObjectList.Create;
  FSQL          := TStringList.Create;
  FExecuted     := False;
  FUnderBDE     := False;
  FHasCursor    := True;
  FHasMacros    := True;
  FHasReference := False;
  FRefName      := EmptyStr;
  FCursorName   := CursorName;
end;

destructor TRepQuery.Destroy;
begin
  FCursorQuery.Free;
  if FQuery <> nil then FQuery.Free;
  if not HasReference and (FBDEQuery <> nil) then FBDEQuery.Free;
  if FParams <> nil then FParams.Free;
  if FSQL <> nil then FSQL.Free;
end;

function TRepQuery.GetSQL: TStrings;
begin
  Result := FSQL;
end;

procedure TRepQuery.SetSQL(Value: TStrings);
begin
  if not HasReference then FSQL.Assign(Value);
end;

function TRepQuery.GetFetchBy: Integer;
begin
  if (not FUnderBDE) and (not FHasReference) then
    if FCursorName = '' then
      Result := FQuery.ReadBuffer
    else
      Result := FCursorQuery.ReadBuffer
  else
    Result := -1;
end;

procedure TRepQuery.SetFetchBy(Value: Integer);
begin
  if (not FUnderBDE) and (not FHasReference) then
    if FCursorName = '' then
      FQuery.ReadBuffer := Value
    else
      FCursorQuery.ReadBuffer := Value;
end;

procedure TRepQuery.SetRefName(Value: String);
begin
  if FRefName = Value then exit;

  if HasReference then
    FBDEQuery := nil;

  FRefName := Value;
  FHasReference := True;

  if SelfPrepared then
  begin
    SelfPrepared := False;
    Prepare;
  end;
end;

// Binds the variables and opens then query
// This method called by the RepEngine when ResetWhen expression is TRUE

procedure TRepQuery.Open;
begin
  Self.Close;
  ApplyParams;
  try
    if FUnderBDE and (not HasReference) then
      if FHasCursor then
        try
          TQuery(FBDEQuery).Prepared := False;
          TQuery(FBDEQuery).Prepared := True;
          FBDEQuery.Open;
        except
          on E: ENoResultSet do
            FHasCursor := False;
          on E: Exception do
            RepEngine.ReportError(6, Self.Name, E.Message, Self, True);
        end
      else
      begin
        TQuery(FBDEQuery).Prepared := False;
        TQuery(FBDEQuery).Prepared := True;
        TQuery(FBDEQuery).ExecSQL;
      end
    else if not HasReference then
    begin
      FQuery.Execute;
      if FCursorName <> '' then FCursorQuery.Execute;
    end;
    FExecuted := True;
  except
    on E: Exception do
      RepEngine.ReportError(6, Self.Name, E.Message, Self, True);
  end;
end;

// Closes the query

procedure TRepQuery.Close;
begin
  if FUnderBDE then
    if not HasReference then
      FBDEQuery.Close
  else
  begin
    if FQuery <> nil then FQuery.Close;
    if FCursorName <> '' then FCursorQuery.Close;
  end;
  FExecuted := False;
end;

// define a variables for query via the report's expressions
procedure TRepQuery.EnumVariables(VarName: String; Expr: TRepExpression;
  ParType: TRepQueryParamType);
begin
  FParams.Add(TRepQueryParam.CreateNew(VarName, Expr, ParType));
end;

procedure TRepQuery.BeginTurn;
begin
  if Processed then exit;
  inherited BeginTurn;
  FNullFields := True;
  FTurned := False;
end;

// fetch a row from the query. This methos will called only when FetchWhen
// expression is TRUE
procedure TRepQuery.Turn;
begin
  if (ResetClause.Addr <> nil) and (ResetClause.Addr.Value = True) then
    Close;

  if not FExecuted then
    Self.Open
  else if not EOF then
    if FUnderBDE then
      FBDEQuery.Next
    else
      if FCursorName = '' then
        FQuery.Next
      else
        FCursorQuery.Next;

  Inc(FRowsFetched);
  FNullFields := EOF;
  RepEngine.FQueryFinished := FNullFields;

  if FModifyParams then
    ApplyQueryParamsToExpr;

  if not FNullFields then
    FTurned := True;
end;

function TRepQuery.GetEOF: Boolean;
begin
  if FUnderBDE then
    Result := FBDEQuery.EOF
  else
    if FCursorName = '' then
      Result := FQuery.Eof
    else
      Result := FCursorQuery.Eof;
end;

procedure TRepQuery.Prepare;
var
  SL: TStringList;
  i: Integer;
  C: TComponent;
  ExprToken: TRepToken;
begin
  if SelfPrepared then exit;

  inherited Prepare;
  FRowsFetched := 0;

  if HasReference then
  begin
    C := VeryGlobalFindComponent(RefName);
    if (C <> nil) and (C is TDataSet) then
    begin
      if FQuery <> nil then FQuery.Free;
      if FBDEQuery <> nil then FBDEQuery.Free;
      FQuery := nil;
      FBDEQuery := nil;

      FBDEQuery := TDataSet(C);
      FUnderBDE := True;
    end;
  end;


  SL := TStringList.Create;
  try
    DescribeMacros(SL);
    for i := 0 to SL.Count-1 do
    begin
      ExprToken := Str2RepToken(SL.Strings[i]);
      if ExprToken.AddrStr = '0' then
        EnumVariables(SL.Strings[i], RepEngine.ExprByToken[ExprToken], ptMacro)
      else
        EnumVariables(SL.Strings[i], TRepExpression(StrToInt(ExprToken.AddrStr)), ptMacro);
    end;

    DescribeParams(FSQL, SL);
    for i := 0 to SL.Count-1 do
      if SL.Strings[i] = AnsiLowerCase(FCursorName) then
        EnumVariables(SL.Strings[i],
                      nil,
                      ptParam)
      else
        EnumVariables(SL.Strings[i],
                      RepEngine.ExprByToken[ToRepToken(SL.Strings[i],
                                                       nil,
                                                       '',
                                                       RepPosition(-1,-1))],
                      ptParam);
  finally
    SL.Free;
  end;

  if FUnderBDE and (not HasReference) then
  begin
    TQuery(FBDEQuery).DatabaseName := Alias;
    TQuery(FBDEQuery).UniDirectional := True;
  end
  else if FQuery <> nil then
  begin
    FQuery.Session := RepEngine.OracleSession;
    FCursorQuery.Session := RepEngine.OracleSession;
  end;

  SelfPrepared := True;
end;

function TRepQuery.GetField(FieldName: String): Variant;
begin
  MakeTurn;
  if FNullFields then
    Result := Null
  else
    if FUnderBDE then
      Result := FBDEQuery.FieldByName(FieldName).Value
    else
      if FCursorName = '' then
        Result := FQuery.Field(FieldName)
      else
        Result := FCursorQuery.Field(FieldName);
end;

function TRepQuery.GetParams(Index: Integer): TRepQueryParam;
begin
  Result := TRepQueryParam(FParams[Index]);
end;

function TRepQuery.GetParamCount: Integer;
begin
  Result := FParams.Count;
end;

// Sets query's variables from the reports's expressions. The references
// between the variables and expressions are have to be eteblished by
// calling the EnumVariables

procedure TRepQuery.ApplyParams;
var
  i: Integer;
  AssumeMacros: Boolean;
begin

  AssumeMacros := FHasMacros and (not HasReference);

  if AssumeMacros then
    BindMacros(FQuery.SQL);

  if not FUnderBDE then
  begin
    FQuery.ClearVariables;
    for i := 0 to ParamCount-1 do
      with Params[i] do
        if FType = ptParam then
          if FExpr = nil then
          begin
            FQuery.DeclareVariable(FName, otCursor);
            FQuery.SetComplexVariable(FName, FCursorQuery);
          end
          else
          begin
            case VarType(FExpr.Value) of
              varSmallint,
              varInteger:  FQuery.DeclareVariable(FName, otInteger);
              varCurrency,
              varSingle,
              varDouble:   FQuery.DeclareVariable(FName, otFloat);
              varDate:     FQuery.DeclareVariable(FName, otDate);
              varString,
              varOleStr:   FQuery.DeclareVariable(FName, otString);
              else
                           FQuery.DeclareVariable(FName, otString);
            end;
            FQuery.SetVariable(FName, FExpr.Value);
          end;
  end
  else
  begin
    if AssumeMacros then
    begin
      TQuery(FBDEQuery).SQL.Assign(FQuery.SQL);
      try
        TQuery(FBDEQuery).Prepare;
      except
        on E: Exception do
          RepEngine.ReportError(6, Self.Name, E.Message, Self, True);
      end;
    end;
    for i := 0 to ParamCount-1 do
      with Params[i] do
        if FType = ptParam then
          if FExpr = nil then
            TQuery(FBDEQuery).ParamByName(FName).DataType := ftCursor
          else
            TQuery(FBDEQuery).ParamByName(FName).Value := FExpr.Value;
  end;
end;

procedure TRepQuery.ApplyQueryParamsToExpr;
var
  i: Integer;
begin
  for i := 0 to ParamCount-1 do
    with Params[i] do
      if FType = ptParam then
        if FUnderBDE then
          FExpr.Value := TQuery(FBDEQuery).ParamByName(FExpr.Name).Value
        else
          FExpr.Value := FQuery.GetVariable(FExpr.Name);
end;

procedure TRepQuery.DescribeMacros(StringList: TStrings);
var
  SrcS: String;
  i: Integer;
  p: Integer;
  Para: String;
begin
  with StringList do
  begin
    Clear;
    for i := 0 to FSQL.Count-1 do
    begin
      SrcS := FSQL.Strings[i];

      while True do
      begin
        p := Pos('&', SrcS);

        if p = 0 then
          break;

        SrcS := Copy(SrcS, p+1, length(SrcS));
        p := Pos('&', SrcS);

        if p = 0 then
          RepEngine.ReportError(3, Self.Name, '', Self, True);

        Para := Copy(SrcS, 1, p-1);
        SrcS := Copy(SrcS, p+1, length(SrcS));
        Add(Para);
      end;
    end;

  end;
end;

procedure TRepQuery.BindMacros(StringList: TStrings);
var
  i: Integer;
  SrcS: String;
  DestS: String;
  ParaIndex: Integer;
  p: Integer;
begin

  ParaIndex := 0;
  FHasMacros := False;
  with StringList do
  begin
    Clear;
    for i := 0 to FSQL.Count-1 do
    begin

      SrcS := FSQL.Strings[i];
      DestS := EmptyStr;
      while True do
      begin
        p := Pos('&', SrcS);
        if p = 0 then
        begin
          DestS := DestS + SrcS;
          break;
        end;

        FHasMacros := True;

        DestS := DestS + Copy(SrcS, 1, p-1);
        SrcS := Copy(SrcS, p+1, Length(SrcS));

        p := Pos('&', SrcS);
        if p = 0 then
          RepEngine.ReportError(3, Self.Name, '', Self, True);

        if p < Length(SrcS) then
          SrcS := Copy(SrcS, p+1, Length(SrcS))
        else
          SrcS := EmptyStr;

        try
          DestS := Trim(DestS + Params[ParaIndex].FExpr.Value);
        except
          RepEngine.ReportError(26,
                                Self.Name,
                                Params[ParaIndex].FExpr.Name,
                                Params[ParaIndex].FExpr, True);
        end;
        Inc(ParaIndex);

      end;
      Add(DestS);
    end;
  end;

end;

function TRepQuery.DoTurn: Boolean;
var
  WhenResult: Boolean;
  ResetResult: Boolean;
begin
  Result := not Processed;

  if Processed then exit;

  if WhenClause.Addr <> nil then
    try
      WhenResult := WhenClause.Addr.Value
    except
      RepEngine.ReportError(73, Self.Name, WhenClause.Addr.Name, WhenClause.Addr, True);
      raise;
    end
  else
    WhenResult := True;

  if ResetClause.Addr <> nil then
    try
      ResetResult := ResetClause.Addr.Value
    except
      RepEngine.ReportError(74, Self.Name, ResetClause.Addr.Name, ResetClause.Addr, True);
      raise;
    end
  else
    ResetResult := False;

  if WhenResult or ResetResult then
    Turn;

  FProcessed := True;
end;

function TRepQuery.GetErrorPosition: TRepPosition;
{
var
  CDA: ^TCDA;
  Size: Word;
  Stmt: HDBIStmt;
}
begin
  Result.Line := -1;
  Result.Col  := -1;

//  CDA := nil;
  
  if not FUnderBDE then
  begin
    if FQuery.ErrorLine > 0 then
    begin
      Result := SQLPosition;
      Result.Line := Result.Line + FQuery.ErrorLine - 1;
      Result.Col  := Result.Col  + FQuery.ErrorPosition - 1;
    end;
  end
  else if not HasReference then
  {
  with TQuery(FBDEQuery) do
    begin
      DbiQAlloc(DBHandle, qrylangSQL, Stmt);
      try
        DbiQPrepare(Stmt, PChar(SQL.Text));
        DbiGetProp(HDBIOBJ(Stmt), stmtNATIVEHNDL, @CDA, 64, Size);
        if CDA <> nil then
          Result.Line := CDA^.peo;
      finally
        DbiQFree(Stmt);
      end;
    end;
  }
end;

{TCustomRepEngine}

constructor TCustomRepEngine.Create(AOwner: TComponent);
begin
  CreateSubreport(AOwner, nil);
end;

constructor TCustomRepEngine.CreateSubreport(AOwner: TComponent;
  AParentReport: TCustomRepEngine);
begin
  inherited Create(AOwner);
  FParentReport     := AParentReport;
  FExprList         := TObjectList.Create;
  FQueryList        := TObjectList.Create;
  FControlList      := TObjectList.Create;
  FTurnableExprList := TObjectList.Create;
  FChildRepList     := TObjectList.Create;
  FParams           := TRepParams.Create(Self);
  FErrorList        := TStringList.Create;
  FExtraHeight      := 0;
  FMaxPages         := 10000;
  FRepStarted       := False;
  FRepFinished      := False;
  FAddParamToList   := False;
  FMultiThreaded    := False;
  FNoWait           := False;
  FPrepared         := False;
  FParsed           := False;
end;

destructor TCustomRepEngine.Destroy;
begin
  Terminate;
  if FExprList <> nil then         FExprList.Free;
  if FQueryList <> nil then        FQueryList.Free;
  if FControlList <> nil then      FControlList.Free;
  if FTurnableExprList <> nil then FTurnableExprList.Free;
  if FParams <> nil then           FParams.Free;
  if FChildRepList <> nil then     FChildRepList.Free;
  ClearErrorList;
  if FErrorList <> nil then FErrorList.Free;
  inherited Destroy;
end;

function TCustomRepEngine.GetCommonRepExpr(Index: Integer): TRepExpression;
begin
  Result := TRepExpression(FExprList[Index]);
end;

function TCustomRepEngine.GetTurnableRepExpr(Index: Integer): TRepExpression;
begin
  Result := TRepExpression(FTurnableExprList[Index]);
end;

function TCustomRepEngine.GetRepQuery(Index: Integer): TRepQuery;
begin
  Result := TRepQuery(FQueryList[Index]);
end;

function TCustomRepEngine.GetRepControl(Index: Integer): TRepControl;
begin
 Result := TRepControl(FControlList[Index]);
end;

function TCustomRepEngine.GetChildRep(Index: Integer): TCustomRepEngine;
begin
  Result := TCustomRepEngine(FChildRepList[Index]);
end;

function TCustomRepEngine.GetCommonRepExprCount: Integer;
begin
  Result := FExprList.Count;
end;

function TCustomRepEngine.GetTurnableRepExprCount: Integer;
begin
  Result := FTurnableExprList.Count;
end;

function TCustomRepEngine.GetRepQueryCount: Integer;
begin
  Result := FQueryList.Count;
end;

function TCustomRepEngine.GetRepControlCount: Integer;
begin
  Result := FControlList.Count;
end;

function TCustomRepEngine.GetChildRepCount: Integer;
begin
  Result := FChildRepList.Count;
end;

function TCustomRepEngine.GetRepExprName(AName: String): TRepExpression;
begin
  Result := GetExprByToken(NameToRepToken(AName));
end;

function TCustomRepEngine.GetExprByToken(AToken: TRepToken): TRepExpression;
var
  i: Integer;
  AName: String;
  O: TRepObject;
begin
  Result := nil;
  AName := AnsiUpperCase(Trim(AToken.Text));
  for i := 0 to CommonRepExprCount-1 do
    if (CommonRepExpr[i] <> nil) and (AnsiUpperCase(CommonRepExpr[i].Name) = AName) then
    begin
      Result := CommonRepExpr[i];
      exit;
    end;


  for i := 0 to TurnableRepExprCount-1 do
    if (TurnableRepExpr[i] <> nil) and (AnsiUpperCase(TurnableRepExpr[i].Name) = AName) then
    begin
      Result := TurnableRepExpr[i];
      Exit;
    end;

  if (Result = nil) and (FParentReport <> nil) then
  begin
    FParentReport.FPreparedObj := FPreparedObj;
    Result := FParentReport.GetExprByToken(AToken);
  end;

  if Result = nil then
  begin
    O := TRepObject.CreateOriginal(Self, EmptyRepToken, EmptyRepToken, AName, AToken.Pos);
    try
      if FPreparedObj <> nil then
        ReportError(7, FPreparedObj.Name, AName, O, True)
      else
        ReportError(7, '<NOT SPECIFIED: caller must specify itself to RepEngine.PreparedObject property>', AName, O, True)
    finally
      O.Free;
    end;
  end;
end;

function TCustomRepEngine.GetRepQueryName(AName: String): TRepQuery;
begin
  Result := GetQueryByToken(NameToRepToken(AName));
end;

function TCustomRepEngine.GetQueryByToken(AToken: TRepToken): TRepQuery;
var
  i: Integer;
  AName: String;
  O: TRepObject;
begin
  Result := nil;

  AName := AnsiUpperCase(Trim(AToken.Text));

  for i := 0 to RepQueryCount-1 do
    if AnsiUpperCase(RepQuery[i].Name) = AName then
    begin
      Result := RepQuery[i];
      break;
    end;

  if (Result = nil) and (FParentReport <> nil) then
  begin
    FParentReport.FPreparedObj := FPreparedObj;
    Result := FParentReport.GetQueryByToken(AToken);
  end;

  if Result = nil then
  begin
    O := TRepObject.CreateOriginal(Self, EmptyRepToken, EmptyRepToken, AName, AToken.Pos);
    try
      ReportError(8, FPreparedObj.Name, AName, O, True);
    finally
      O.Free;
    end;
  end;
end;

function TCustomRepEngine.GetRepControlName(AName: String): TRepControl;
begin
  Result := GetControlByToken(NameToRepToken(AName));
end;

function TCustomRepEngine.GetControlByToken(AToken: TRepToken): TRepControl;
var
  i: Integer;
  AName: String;
  O: TRepObject;
begin
  Result := nil;

  AName := AnsiUpperCase(Trim(AToken.Text));

  for i := 0 to RepControlCount-1 do
    if AnsiUpperCase(RepControl[i].Name) = AName then
    begin
      Result := RepControl[i];
      break;
    end;

  if (Result = nil) and (FParentReport <> nil) then
  begin
    FParentReport.FPreparedObj := FPreparedObj;
    Result := FParentReport.GetControlByToken(AToken);
  end;


  if Result = nil then
  begin
    O := TRepObject.CreateOriginal(Self, EmptyRepToken, EmptyRepToken, AName, AToken.Pos);
    try
      ReportError(9, FPreparedObj.Name, AName, O, True);
    finally
      O.Free;
    end;
  end;
end;

function TCustomRepEngine.GetChildRepName(AToken: TRepToken): TCustomRepEngine;
var
  i: Integer;
  AName: String;
  O: TRepObject;
begin
  Result := nil;

  AName := AnsiUpperCase(Trim(AToken.Text));

  for i := 0 to ChildRepCount-1 do
    if AnsiUpperCase(ChildRep[i].Name) = AName then
    begin
      Result := ChildRep[i];
      break;
    end;

  if (Result = nil) and (FParentReport <> nil) then
  begin
    FParentReport.FPreparedObj := FPreparedObj;
    Result := FParentReport.GetChildRepName(AToken);
  end;


  if Result = nil then
  begin
    O := TRepObject.CreateOriginal(Self, EmptyRepToken, EmptyRepToken, AName, AToken.Pos);
    try
      ReportError(10, FPreparedObj.Name, AName, O, True);
    finally
      O.Free;
    end;
  end;
end;

procedure TCustomRepEngine.RefreshCtrlList(ARepControl: TRepControl);
begin
  if (ARepControl = nil) or (FControlList.IndexOf(ARepControl) <> -1) then
  begin
    FControlList.Sort(CompareRepCtrlsOrder);
  end;
end;

procedure TCustomRepEngine.ReportError(ErrorCode: Word;
                                       ErrorObjectName, ExtraData: String;
                                       RepObject: TRepObject; RaiseException: Boolean);
var
  S: String;
  P: ^TRepPosition;
  Str: String;
  l: Integer;
  ErrCnt: Integer;
  OraError: Boolean;
begin
  if FParentReport <> nil then
    FParentReport.ReportError(ErrorCode, ErrorObjectName, ExtraData, RepObject,
                              RaiseException);

  Str := 'REP-'+FormatFloat('00000', ErrorCode)+': '+
         FmtLoadStr(ErrorCode + 10000, [ErrorObjectName, ExtraData]);

  OraError := (Pos('ORA', Str) > 0);

  ErrCnt := 0;
  while True do
  begin
    l := Pos(#10, Str);
    if l > 0 then
      S := Copy(Str, 1, l-1)
    else
      S := Str;

    if ErrCnt <> 0 then
      S := '    '+S;

    GetMem(P, SizeOf(TRepPosition));
    with P^ do
    begin
      if RepObject <> nil then
        if (RepObject is TRepQuery) and OraError then
          with TRepQuery(RepObject) do
          begin
            Line := ErrorPosition.Line;
            Col  := ErrorPosition.Col;
          end
        else
        begin
          Line := RepObject.ScriptPos.Line;
          Col  := RepObject.ScriptPos.Col;
        end
      else
      begin
        Line := 0;
        Col  := 0;
      end;
      S := Format('%d/%d ==> %s', [Line+1, Col+1, S]);
    end;

    FErrorList.AddObject(S, TObject(P));

    if l > 0 then
      Str := Copy(Str, l+1, Length(Str))
    else
      break;

    Inc(ErrCnt);
  end;

  if FReportThread <> nil then
    FReportThread.Terminate;

  if RaiseException then
    raise EReportTerminated.Create(S, ErrorCode);
end;

procedure TCustomRepEngine.SetPrepared(Value: Boolean);
begin
  if FPrepared = Value then exit;

  if FReportThread <> nil then
    ReportError(23, Self.Name, '', nil, True);


  if not Value then
    ResetReport
  else
    PrepareReport(True);

  FPrepared := Value;
end;

procedure TCustomRepEngine.ResetReport;
begin
  ClearErrorList;
  FQueryList.ClearObjects;
  FExprList.ClearObjects;
  FTurnableExprList.ClearObjects;
  FControlList.ClearObjects;
  FChildRepList.ClearObjects;
  FExtraHeight := 0;
  FRepStarted   := True;
  FRepFinished  := False;
  FCurPageHeight := 0;
  FPageState := psStarting;
  FPageCount := 0;
  FExtraHeightUsed := FExtraHeight;
  FPrepared := False;
  FParsed   := False;
end;

procedure TCustomRepEngine.PrepareReport(ParseIt: Boolean);
begin
  if FPrepared then exit;

  if ParseIt and (not FParsed) then
  begin
    ResetReport;
    Parse;
  end;
  ParseSubreports;
  FParams.Bind;
  PrepareList(FQueryList);
  PrepareList(FExprList);
  PrepareList(FTurnableExprList);
  PrepareList(FControlList);
  PrepareSubreports;

  if (FParentReport <> nil) and (FWhenClause.Text <> EmptyStr) then
    FWhenClause.Addr := FParentReport.ExprByToken[FWhenClause];

  FPrepared := True;
end;

procedure TCustomRepEngine.GetParams;
begin
  if FPrepared then exit;
  try
    try
      FParams.Clear;
      ResetReport;
      FAddParamToList := True;
      Parse;
      FParsed := True;
//      ResetReport;
    finally
      FAddParamToList := False;
    end;
  except
    on EReportTerminated do
    begin
      if csDesigning in ComponentState then
        OpenErrors;
    end
    else
      raise;
  end;
end;

procedure TCustomRepEngine.InternalRun;
begin
  Prepare;

  FRepStarted   := True;
  FRepFinished  := False;

  if FParentReport <> nil then
    with FParentReport do
    begin
      Self.FCurPageHeight := FCurPageHeight - (FExtraHeight - FExtraHeightUsed);
      Self.FPageState := FPageState;
      Self.FPageCount := FPageCount;
    end
  else
  begin
    FCurPageHeight := 0;
    FPageState := psStarting;
    FPageCount := 0;
  end;

  FExtraHeightUsed := FExtraHeight;

  try
    ResetExpressions;
    while True do
    begin
      WaitForSubreports;
      ProcessQueries;
      EvalExpressions;
      RunSubreports;
      while not PrintControls do;
      ProcessEndTurn;
      if RepFinished then
      begin
        if (FCurPageHeight <> 0) or (FExtraHeight <> FExtraHeightUsed) then
          Inc(FPageCount);
        Break;
      end;
      if FReportThread.Terminated then break;
    end;
  finally
    UnPrepare;
  end;

  if FParentReport <> nil then
  begin
    FParentReport.FCurPageHeight := FCurPageHeight - (FExtraHeight - FExtraHeightUsed) - 1;
    FParentReport.FPageState := FPageState;
    FParentReport.FPageCount := FPageCount;
  end

end;

procedure TCustomRepEngine.Run;
begin

  if FReportThread <> nil then
    ReportError(23, Self.Name, '', nil, True);

  FReportThread := TReportThread.Create(Self);
  with FReportThread do
  begin
    FreeOnTerminate := True;
    OnTerminate := ReportTerminated;
    Resume;
    if (not FMultiThreaded) or ((FParentReport <> nil) and (not FNoWait)) then
      WaitForTerminate;
  end;
end;

procedure TCustomRepEngine.ReportTerminated(Sender: TObject);
begin
  if Assigned(FOnTerminate) then
    if FReportThread.Terminated then
      FOnTerminate(Self, ttAbort)
    else if FErrorList.Count <> 0 then
      FOnTerminate(Self, ttError)
    else
      FOnTerminate(Self, ttNormal);

  FReportThread := nil;
end;

procedure TCustomRepEngine.PrepareList(AList: TObjectList);
var
  i: Integer;
begin
  i := 0;
  while i < AList.Count do
  begin

    if (FReportThread <> nil) and (FReportThread.Terminated) then exit;

    FListFreshed := True;
    with  TRepObject(AList.Items[i]) do
    begin
      FPreparedObj := TRepObject(AList.Items[i]);
      Prepare;
    end;
    if not FListFreshed then
      i := 0
    else
      Inc(i);
  end;
end;

procedure TCustomRepEngine.PrepareSubreports;
var
  i: Integer;
begin
  if (FReportThread <> nil) and (FReportThread.Terminated) then exit;
  for i := 0 to ChildRepCount-1 do
    ChildRep[i].PrepareReport(False);
end;

procedure TCustomRepEngine.ParseSubreports;
var
  i: Integer;
begin
  for i := 0 to ChildRepCount-1 do
    with ChildRep[i] do
    begin
       ResetReport;
       FChildRepIndex := i;
       if FReportThread <> nil then
         Self.FReportThread.SyncDoBeforeSubRepParsed
       else
         Self.DoBeforeSubRepParsed(i);
       Parse;
    end;
end;

function TCustomRepEngine.ProcessQueries: Boolean;
var
  i: Integer;
  EOFs: Boolean;
begin
  Result := True;
  EOFs := True;
  for i := 0 to RepQueryCount-1 do
    with RepQuery[i] do
    begin
      BeginTurn;
      MakeTurn;
      Result := (Result and (not FTurned));
      EOFs := EOFs and FQueryFinished;
   end;
  FRepFinished := Result or EOFs;
end;

procedure TCustomRepEngine.EvalExpressions;
var
  i: Integer;
begin
  for i := 0 to TurnableRepExprCount-1 do
    with TurnableRepExpr[i] do
    begin
      BeginTurn;
      MakeTurn;
    end;
end;

procedure TCustomRepEngine.ResetExpressions;
var
  i: Integer;
begin
  for i := 0 to TurnableRepExprCount-1 do
    with TurnableRepExpr[i] do
      if (ResetClause.Text <> EmptyStr) and (not Processed)then
        Value := ResetClause.Addr.Value;
end;

procedure TCustomRepEngine.RunSubreports;
var
  i: Integer;
begin
  for i := 0 to ChildRepCount-1 do
    with ChildRep[i] do
    begin
      RepProcessed := False;
      MultiThreaded := Self.MultiThreaded;
      MakeTurn;
    end;
end;

procedure TCustomRepEngine.WaitForSubreports;
var
  i: Integer;
begin
  for i := 0 to ChildRepCount-1 do
    with ChildRep[i] do
      if not NoWait then
        WaitForTerminate;
end;

function TCustomRepEngine.PrintControls: Boolean;
var
  i: Integer;
begin
  Result := True;
  for i := 0 to RepControlCount-1 do
    with RepControl[i] do
      if (WhenClause.Addr = nil) or
        ((WhenClause.Addr <> nil) and (WhenClause.Addr.Value = True)) then
        if (FCurPageHeight + Height > Self.PageHeight + FExtraHeight) and
          (not FAlwaysOnPage) then
        begin
          if FPageState = psStarted then
          begin
            FReportThread.SignalForPage;
            Result := False;
            FPageState := psFinishing;
            exit;
          end
          else if (FPageState = psFinishing) or (FPageState = psSafeFinishing) then
          begin
            Result := False;
            FPageState := psStarting;
            Skip(Self.PageHeight - FCurPageHeight + FExtraHeight - FExtraHeightUsed,
                 True);
            FCurPageHeight := 0;
            FExtraHeightUsed := FExtraHeight;
            Inc(FPageCount);
            exit;
          end
          else
            ReportError(64, Self.Name, '', nil, True);
        end
        else
        begin
          BeginTurn;
          if MakeTurn then
          begin
             if FAlwaysOnPage then
               Inc(FExtraHeightUsed, Height)
             else
               Inc(FCurPageHeight, Height);

             if FinishesPage {and (not FRepFinished)} then
               if DisableFooter then
               begin
                 Result := False;
                 Skip(Self.PageHeight - FCurPageHeight + FExtraHeight - FExtraHeightUsed, False);
                 FCurPageHeight := Self.PageHeight;
                 FPageState := psSafeFinishing;
                 exit;
               end
               else
               begin
                 Result := False;
                 Skip(Self.PageHeight - FCurPageHeight + FExtraHeight, False);
                 FCurPageHeight := Self.PageHeight + FExtraHeight;
                 FPageState := psFinishing;
                 exit;
               end
          end;
        end;
  FPageState := psStarted;
end;

procedure TCustomRepEngine.ProcessEndTurn;
var
  i: Integer;
begin
  for i := 0 to TurnableRepExprCount-1 do
    with TurnableRepExpr[i] do
      if (not ResetBefore) or ResetAfter then
        if Reset then
          if not ResetAfter then
            Turn;

  for i := 0 to TurnableRepExprCount-1 do
    TurnableRepExpr[i].EndTurn;
  for i := 0 to RepQueryCount-1 do
    RepQuery[i].EndTurn;
  for i := 0 to RepControlCount-1 do
    RepControl[i].EndTurn;

  FRepStarted := False;
end;

function TCustomRepEngine.MakeTurn: Boolean;
begin
  Result := False;
  if (FWhenClause.Text = EmptyStr) or
    ((FWhenClause.Text <> EmptyStr) and (FWhenClause.Addr.Value = True)) then
  begin
    Run;
    FRepProcessed := True;
    Result := True;
  end;
end;

function TCustomRepEngine.AddExpression(Expr: TRepExpression): TRepExpression;
begin
  Result := Expr;

  if Expr = nil then exit;

  if Expr.Turnable then
    FTurnableExprList.Add(Expr)
  else
    FExprList.Add(Expr);
end;

function TCustomRepEngine.AddQuery(AQuery: TRepQuery): TRepQuery;
begin
  if AQuery <> nil then
    FQueryList.Add(AQuery);

  Result := AQuery;
end;

function TCustomRepEngine.AddControl(AControl: TRepControl): TRepControl;
begin
  if AControl <> nil then
    FControlList.Add(AControl);
  Result := AControl;
end;

function TCustomRepEngine.AddSubreport(ASubreport: TCustomRepEngine):
  TCustomRepEngine;
begin
  if ASubreport <> nil then
    FChildRepList.Add(ASubreport);
  Result := ASubreport;
end;

procedure TCustomRepEngine.RefreshList;
begin
  FListFreshed := False;
end;

procedure TCustomRepEngine.Prepare;
begin
  if FPrepared then exit;
  PrepareReport(True);
  FPrepared := True;
end;

procedure TCustomRepEngine.UnPrepare;
var
  i: Integer;
begin
  for i := 0 to FQueryList.Count-1 do
    TRepQuery(FQueryList.Items[i]).Close;
end;

procedure TCustomRepEngine.SetParamsList(Value: TRepParams);
begin
  if FParams = nil then
    FParams := TRepParams.Create(Self);

  FParams.Assign(Value);
end;

procedure TCustomRepEngine.WaitForTerminate;
begin
  if FReportThread <> nil then
    FReportThread.WaitFor;
end;

procedure TCustomRepEngine.Terminate;
begin
  if FReportThread <> nil then
  begin
    FReportThread.Terminate;
    FReportThread.WaitFor;
  end;
end;

procedure TCustomRepEngine.WaitForNextPage;
begin
  if FReportThread <> nil then
    FReportThread.WaitForPage;
end;

function TCustomRepEngine.GetTerminated: Boolean;
begin
  Result := FReportThread = nil;
end;

function TCustomRepEngine.GetHasErrors: Boolean;
begin
  if FErrorList <> nil then
    Result := FErrorList.Count <> 0
  else
    Result := False
end;

function TCustomRepEngine.GetSysAddress(AddressType: Word): Pointer;
begin
  case AddRessType of
    SYS_ADDR_CURPAGEHEIGHT:
      Result := @FCurPageHeight;
    SYS_ADDR_EXTRA_HEIGHT:
      Result := @FExtraHeight;
    SYS_ADDR_EXTRA_USED:
      Result := @FExtraHeightUsed;
    else
      Result := nil;
  end;
end;

procedure TCustomRepEngine.ClearErrorList;
var
  i: Integer;
  P: ^TRepPosition;
begin
  if FErrorList = nil then exit;

  for i := 0 to FErrorList.Count-1 do
    if FErrorList.Objects[i] <> nil then
    begin
      P := Pointer(FErrorList.Objects[i]);
      FreeMem(P);
    end;

  FErrorList.Clear;
end;

function TCustomRepEngine.RepVersion: TRepVersion;
begin
  Result.Major := 2;
  Result.Minor := 4;
  Result.Release := 8;
  Result.Comment := 'Beta';
end;

{TPageEndExpression}

constructor TPageEndExpression.Create(RepEngine: TCustomRepEngine;
                                      Pos: TRepPosition);
begin
  inherited CreateOriginal(RepEngine, EmptyRepToken, EmptyRepToken,
                           'PageHeaderSwitch', Pos);
end;

function TPageEndExpression.GetAsVariant: Variant;
begin
  Result := (RepEngine.FPageState = psFinishing);
end;

{TPageStartExpression}

constructor TPageStartExpression.Create(RepEngine: TCustomRepEngine;
                                        Pos: TRepPosition);
begin
  inherited CreateOriginal(RepEngine, EmptyRepToken, EmptyRepToken,
                           'PageFooterSwitch', Pos);
end;

function TPageStartExpression.GetAsVariant: Variant;
begin
  Result := (RepEngine.FPageState = psStarting);
end;

{TRepEndExpression}

constructor TRepEndExpression.Create(RepEngine: TCustomRepEngine;
                                     Pos: TRepPosition);
begin
  inherited CreateOriginal(RepEngine, EmptyRepToken, EmptyRepToken,
                           'REP_END_EXPR', Pos);
end;

function TRepEndExpression.GetAsVariant: Variant;
begin
  Result := RepEngine.RepFinished;
end;

{TRepStartExpression}

constructor TRepStartExpression.Create(RepEngine: TCustomRepEngine;
                                       Pos: TRepPosition);
begin
  inherited CreateOriginal(RepEngine, EmptyRepToken, EmptyRepToken,
                           'REP_START_EXPR', Pos);
end;

function TRepStartExpression.GetAsVariant: Variant;
begin
  Result := RepEngine.RepStarted;
end;

{TPageCountExpression}

constructor TPageCountExpression.Create(RepEngine: TCustomRepEngine;
                                        Pos: TRepPosition);
begin
  inherited CreateOriginal(RepEngine, EmptyRepToken, EmptyRepToken,
                           'REP_PAGE_COUNT', Pos);
end;

function TPageCountExpression.GetAsVariant: Variant;
begin
  Result := RepEngine.PageCount;
end;

{TRepParam}

constructor TRepParam.Create(AParamList: TRepParams; AName: String; Pos: TRepPosition);
begin
  FParamName := AName;
  FParamList := AParamList;
  FValue     := Null;
  FPosition  := Pos;
  AParamList.AddParam(Self);
end;

destructor TRepParam.Destroy;
begin
end;

procedure TRepParam.Assign(Source: TPersistent);
begin
  with TRepParam(Source) do
  begin
    Self.FParamName := ParamName;
    Self.Value := Value;
    Self.FInputMask := InputMask;
  end;
end;

procedure TRepParam.SetValue(AValue: Variant);
begin
  FValue := AValue;
  FParamList.FRepEngine.Prepared := False;
end;

function TRepParam.GetDataSize: Integer;
begin
  Result := 0;
  case VarType(FValue) of
    varSmallInt: Result := sizeOf(SmallInt);
    varInteger: Result := sizeOf(Integer);
    varSingle : Result := sizeOf(Single);
    varDouble, varDate: Result := sizeOf(Double);
    varCurrency: Result := sizeOf(Currency);
    varBoolean: Result := sizeOf(WordBool);
    varString: Result := Length(FValue);
  end;
end;

procedure TRepParam.GetData(Buffer: Pointer);
begin
  case VarType(FValue) of
    varSmallInt: SmallInt(Buffer^) := FValue;
    varInteger: Integer(Buffer^) := FValue;
    varSingle : Single(Buffer^) := FValue;
    varDouble, varDate: Double(Buffer^) := FValue;
    varCurrency: Currency(Buffer^) := FValue;
    varBoolean: WordBool(Buffer^) := FValue;
    varString: StrMove(Buffer, PChar(String(FValue)), Length(FValue)+1);
  end;
end;

procedure TRepParam.SetData(Buffer: Pointer; DataType: Word);
begin
  case DataType of
    varSmallInt: Value := SmallInt(Buffer^);
    varInteger: Value := Integer(Buffer^);
    varSingle : Value := Single(Buffer^);
    varDouble, varDate: Value := Double(Buffer^);
    varCurrency: Value := Currency(Buffer^);
    varBoolean: Value := WordBool(Buffer^);
    varString: Value := StrPas(Buffer);
  end;
end;

{TRepParams}

constructor TRepParams.Create(ARepEngine: TCustomRepEngine);
begin
  FList := TObjectList.Create;
  FRepEngine := ARepEngine;
end;

destructor TRepParams.Destroy;
begin
  FList.Free;
end;


function TRepParams.GetParamValue(const ParamName: String): Variant;
var
  P: TRepParam;
begin
  P := ParamByName(ParamName);
  if P <> nil then
    Result := P.Value
  else
    FRepEngine.ReportError(11, ParamName, '', nil, True);
end;

procedure TRepParams.SetParamValue(const ParamName: String; Value: Variant);
var
  P: TRepParam;
begin
  P := ParamByName(ParamName);
  if P <> nil then
    P.Value := Value
  else
    FRepEngine.ReportError(11, ParamName, '', nil, True);
end;

procedure TRepParams.AddParam(Value: TRepParam);
begin

  if Value = nil then exit;

  if FList.IndexOf(Value) = -1 then
    FList.Add(Value)
  else
    FRepEngine.ReportError(12, Value.FParamName, '', nil, True);
end;

procedure TRepParams.RemoveParam(Value: TRepParam);
begin
  if FList.IndexOf(Value) = -1 then
    FRepEngine.ReportError(13, Value.FParamName, '', nil, True);

  FList.Remove(Value);
end;

function TRepParams.CreateParam(AName: String; Pos: TRepPosition): TRepParam;
begin
  Result := TRepParam.Create(Self, AName, Pos);
end;

function TRepParams.ParamByName(const AParamName: String): TRepParam;
var
  i: Integer;
begin
  Result := nil;

  for i := 0 to FList.Count do
    with TRepParam(FList.Items[i]) do
      if AnsiUpperCase(ParamName) = AnsiUpperCase(AParamName) then
      begin
        Result := TRepParam(FList.Items[i]);
        break;
      end;
end;

procedure TRepParams.Bind;
var
  i: Integer;
begin
  for i := 0 to FList.Count-1 do
    with TRepParam(FList.Items[i]) do
      FRepEngine.FExprList.Add(TRepConstExpression.Create(FRepEngine,
        Value, AnsiUpperCase(ParamName), FPosition));
end;

procedure TRepParams.Clear;
begin
  FList.ClearObjects;
end;

function TRepParams.GetParamCount: Integer;
begin
  Result := FList.Count;
end;

function TRepParams.GetParam(Index: Integer): TRepParam;
begin
  Result := TRepParam(FList.Items[Index]);
end;

procedure TRepParams.Assign(Source: TPersistent);
var
  i: Integer;
  p: TRepParam;
begin
  FRepEngine := TRepParams(Source).FRepEngine;
  Clear;
  for i := 0 to TRepParams(Source).ParamCount-1 do
  begin
    p := TRepParam.Create(Self, '', RepPosition(-1,-1));
    p.Assign(TRepParams(Source).Param[i]);
  end;
end;

procedure TRepParams.DefineProperties(Filer: TFiler);
begin
  inherited DefineProperties(Filer);
  Filer.DefineBinaryProperty('Data', ReadBinaryData, WriteBinaryData, True);
end;

procedure TRepParams.ReadBinaryData(Stream: TStream);
var
  i: Integer;
  NumItems: Integer;
  Buffer: array[0..4096] of char;
  Temp: Integer;
  TmpWord: Word;
begin
  with Stream do
  begin
    ReadBuffer(NumItems, sizeOf(NumItems));
    for i := 0 to NumItems - 1 do
      with TRepParam.Create(Self, EmptyStr, RepPosition(-1,-1)) do
      begin
        ReadBuffer(Temp, SizeOf(Temp));
        SetLength(FParamName, Temp);
        ReadBuffer(PChar(FParamName)^, Temp);
        ReadBuffer(TmpWord, sizeOf(TmpWord));

        ReadBuffer(Temp, sizeOf(Temp));
        ReadBuffer(Buffer, Temp);
        Buffer[Temp] := #0;
        SetData(@Buffer, TmpWord);
      end;
  end;

end;

procedure TRepParams.WriteBinaryData(Stream: TStream);
var
  i: Integer;
  Temp: Integer;
  TmpWord: Word;
  Buffer: array[0..4096] of char;
begin

  with Stream do
  begin
    i := ParamCount;
    WriteBuffer(i, SizeOf(i));
    for i := 0 to ParamCount-1 do
      with Param[i] do
      begin
        Temp := Length(ParamName);
        WriteBuffer(Temp, SizeOf(Temp));

        WriteBuffer(PChar(ParamName)^, Temp);

        TmpWord := TVarData(Value).VType;
        WriteBuffer(TmpWord, sizeOf(TmpWord));

        Temp := GetDataSize;
        WriteBuffer(Temp, SizeOf(Temp));

        GetData(@Buffer);
        WriteBuffer(Buffer, Temp);
      end;
  end;

end;

function VeryGlobalFindComponent(Name: String): TComponent;
var
  I: Integer;
begin
  for I := 0 to Screen.FormCount - 1 do
  begin
    Result := Screen.Forms[I];
    if CompareText(Name, Result.Name) = 0 then Exit;
    Result := Result.FindComponent(Name);
    if Result <> nil then exit;
  end;
  for I := 0 to Screen.DataModuleCount - 1 do
  begin
    Result := Screen.DataModules[I];
    if CompareText(Name, Result.Name) = 0 then Exit;
    Result := Result.FindComponent(Name);
    if Result <> nil then exit;
  end;
  Result := nil;
end;


{TReportThread}

constructor TReportThread.Create(RepEngine: TCustomRepEngine);
begin
  FRepEngine := RepEngine;
  FSemaphore := CreateSemaphore(nil, 0, FRepEngine.MaxPages, nil);
  inherited Create(True);
end;

destructor TReportThread.Destroy;
begin
  CloseHandle(FSemaphore);
  FSemaphore := 0;
  inherited Destroy;
end;

procedure TReportThread.SignalForPage;
begin
  ReleaseSemaphore(FSemaphore, 1, nil);
  if Assigned(FRepEngine.FOnNextPage) then
    Synchronize(DoOnNextPage);
end;

procedure TReportThread.WaitForPage;
var
  H: array[0..1] of integer;
  Msg: Tmsg;
begin
  H[0] := Self.Handle;
  H[1] := FSemaphore;
  while MsgWaitForMultipleObjects(2, H, False, INFINITE,
    QS_SENDMESSAGE) = WAIT_OBJECT_0 do
    PeekMessage(Msg, 0, 0, 0, PM_NOREMOVE);
end;

procedure TReportThread.Execute;
begin
  try
    FRepEngine.InternalRun;
  except
    on EReportTerminated do
      Terminate;
    on E: Exception do
      FRepEngine.ReportError(71, FRepEngine.Name, E.Message, nil, False);
  end;
end;

procedure TReportThread.DoOnNextPage;
begin
  if FRepEngine.PageCount = 1 then
    DoOnReportStart;

  FRepEngine.FOnNextPage(FRepEngine);
end;

procedure TReportThread.DoOnReportStart;
begin
  if Assigned(FRepEngine.FOnStart) then
    FRepEngine.FOnStart(FRepEngine);
end;

procedure TReportThread.SyncDoBeforeSubRepParsed;
begin
  with FRepEngine do
    DoBeforeSubRepParsed(FChildRepIndex);
end;

function EmptyRepToken: TRepToken;
begin
  Result.Text := '';
  Result.Addr := nil;
  Result.Pos.Line := -1;
  Result.Pos.Col  := -1;
end;

function Str2RepToken(Str: String): TRepToken;

  function NextToken: String;
  var
    l: Integer;
  begin
    l := Pos(',', Str);
    if l > 0 then
    begin
      Result := Trim(Copy(Str, 1, l-1));
      Str := Copy(Str, l+1, Length(Str));
    end
    else
    begin
      Result := Trim(Str);
      Str := '';
    end;
  end;

begin
  Result.Text     := NextToken;
  Result.Addr     := TRepExpression(StrToInt(NextToken));
  Result.Pos.Line := StrToInt(NextToken);
  Result.Pos.Col  := StrToInt(NextToken);
  Result.AddrStr  := IntToStr(Integer(Result.Addr));
end;

function ToRepToken(Text: String;
                    Addr: TRepExpression;
                    AddrStr: String;
                    Pos: TRepPosition): TRepToken;
begin
  Result.Text := Text;
  Result.Addr := Addr;
  Result.AddrStr := AddrStr;
  Result.Pos := Pos;
end;

function NameToRepToken(AName: String): TRepToken;
begin
  Result := ToRepToken(AName, nil, '', RepPosition(-1, -1));
end;

end.

