{**
*
*         Script Reporter Component library
*         Script language compiler
*          Vlad Ilushchenko 1998-99
*
*
**}

unit Reporter;

interface

uses RepKrnl, RepExpr, Tokenizer, Graphics, Classes, SysUtils,
  ObjList;

type
  TCustomScriptRep = class;

  TArgumentKind = (akLeft, akRight, akSet);
  TArgumentKinds = set of TArgumentKind;

  TSubReportParseEvent = procedure(Sender: TCustomScriptRep; SubReport: TCustomScriptRep) of object;

  EDuplicatePlugInName = class(Exception) end;

  TCustomScriptRep = class(TCustomRepEngine)
  private
    FScriptFileName: String;
    FOutputFileName: String;
    FProcessed: Boolean;
    FTitle: String;
    FFuncList: TList;
    FAppendFile: Boolean;
    FParentOutput: Boolean;
    FOutStream: TFileStream;
    FAlignPages: Boolean;
    FBeforeSubReportParsed: TSubReportParseEvent;
    FOnLineProcessed: TLineProcessedEvent;
    procedure SetScriptFile(Value: String);
    procedure SetProcessed(Value: Boolean);
  protected
    procedure Parse; override;
    procedure Print(AText: string; AFont: TFont); override;
    procedure Skip(Units: Integer; ToNewPage: Boolean); override;
    procedure Prepare; override;
    procedure UnPrepare; override;
    procedure ParseScript(Parser: TStreamTokenizer);
    procedure SetPrepared(Value: Boolean); override;
    function GetStream: TFileStream;
    procedure DoBeforeSubRepParsed(Index: Integer); override;
  public
    constructor Create(AOwner: TComponent); override;
    constructor CreateSubreport(AOwner: TComponent;
      AParentReport: TCustomRepEngine); override;
    destructor Destroy; override;

    procedure ScriptError(ErrorCode: Integer; Line, Col: Integer);
    procedure OpenErrors; override;
    procedure OpenCompiler; override;
    property ScriptFile: String read FScriptFileName write SetScriptFile;
    property OutputFile: String read FOutputFileName write FOutputFileName;
    property Processed: Boolean read FProcessed write SetProcessed stored False;
    property Title: String read FTitle write FTitle;
    property AppendFile: Boolean read FAppendFile write FAppendFile;
    property ParentOutput: Boolean read FParentOutput write FParentOutput;
    property AlignPages: Boolean read FAlignPages write FAlignPages;
    property BeforeSubReportParsed: TSubReportParseEvent read FBeforeSubReportParsed
                                                         write FBeforeSubReportParsed;
    property OnLineProcessed: TLineProcessedEvent read FOnLineProcessed write FOnLineProcessed;
  end;

  EScripFileEmpty = class(Exception) end;

procedure AddPlugIn(FuncName: String;
                    Args: TArgumentKinds;
                    ExprClass: TRepExpressionClass; IsSeparate: Boolean);

implementation

uses RepCtrl, Dialogs, Controls, RepErrorsForm, ExprParser, DebuggerForm;

type
  TControlOrder = (coHighest, coHigh, coNormal, coLow, coDefault);

  TExprDesc = class
  public
    ReservedName: String;
    ArgKinds: TArgumentKinds;
    ExprClass: TRepExpressionClass;
    AssumeSeparate: Boolean;

    constructor Create(AName: String; AnArgs: TArgumentKinds;
      AnExprClass: TRepExpressionClass);
  end;

  TConstType = (ctBad, ctDigit, ctString, ctBool);
  TForwardKind = (fkQuery, fkExpression, fkControl);

  TForwardDef = class
  public
    Name: String;
    Kind: TForwardKind;
    constructor Create(AName: String; AKind: TForwardKind);
  end;

var
  PlugInList: TObjectList;

procedure InitPlugInList;
begin
  if PlugInList = nil then
    PlugInList := TObjectList.Create;
end;

function AddrStr(Value: TRepToken): String;
begin
  Result := IntToStr(Integer(Value.Addr));
end;

const
  NHighControlRange = 10000;
  NMidControlRange = 5000;
  NLowControlRange = 1000;
  NLowestControlRange = 100;

{TExprDesc}

constructor TExprDesc.Create;
begin
  inherited Create;
  ReservedName   := AName;
  ArgKinds       := AnArgs;
  ExprClass      := AnExprClass;
  AssumeSeparate := False;
end;

{TCustomScriptRep}

constructor TCustomScriptRep.Create(AOwner: TComponent);
begin
  CreateSubreport(AOwner, nil);
  FProcessed := False;
  FFuncList := TList.Create;
  FParentOutput := False;
  FAlignPages := True;
end;

constructor TCustomScriptRep.CreateSubreport(AOwner: TComponent;
  AParentReport: TCustomRepEngine);
begin
  inherited CreateSubreport(AOwner, AParentReport);
  FProcessed := False;
  FFuncList := TList.Create;
  FParentOutput := False;
end;

destructor TCustomScriptRep.Destroy;
begin
  if FFuncList <> nil then FFuncList.Free;
  inherited Destroy;
end;

procedure TCustomScriptRep.SetScriptFile(Value: String);
begin
  if FScriptFileName <> Value then
  begin
    FScriptFileName := Value;
    Processed := False;
  end;
end;

procedure TCustomScriptRep.SetProcessed(Value: Boolean);
begin
  if FProcessed <> Value then
  begin
    if Value then
      Run
    else
      if (csDesigning in ComponentState) and (not Terminated) then
      begin
        if MessageDlg('Reporter is now executing! Terminate?',
          mtWarning, [mbYes, mbNo], 0) = mrYes then
          Terminate;
      end
      else
        Terminate;
  end;
  FProcessed := Value;
end;

procedure TCustomScriptRep.Prepare;
begin
  inherited Prepare;

  if (ParentReport = nil) or (not FParentOutput) then
    try
      if not FAppendFile then
      begin
        FOutStream := TFileStream.Create(FOutputFileName, fmCreate);
        FOutStream.Free;
      end;
      FOutStream := TFileStream.Create(FOutputFileName, fmOpenReadWrite or
        fmShareDenyNone);
      FOutStream.Position := FOutStream.Size;
    except
      ScriptError(63, -1, -1);
      raise;
    end
  else
    FOutStream := TCustomScriptRep(ParentReport).GetStream;

end;

procedure TCustomScriptRep.UnPrepare;
begin
  inherited UnPrepare;
  if (ParentReport = nil) or (not FParentOutput) then
    FOutStream.Free;
end;

function TCustomScriptRep.GetStream: TFileStream;
begin
  if (ParentReport = nil) or (not FParentOutput) then
    Result := FOutStream
  else
    Result := TCustomScriptRep(ParentReport).GetStream;
end;

procedure TCustomScriptRep.Print(AText: String; AFont: TFont);
begin
  AText := AText + #13+#10;
  try
    FOutStream.WriteBuffer(Pointer(AText)^, Length(AText));
  except
    ReportError(27, FOutputFileName, Self.Name, nil, True);
  end;
end;

procedure TCustomScriptRep.Skip(Units: Integer; ToNewPage: Boolean);
var
  i: Integer;
begin
  if FAlignPages or (not ToNewPage) then
    for i := 0 to Units-1 do
      Print('', nil)
  else
    Print(Chr(12), nil);
end;

procedure TCustomScriptRep.Parse;
var
  S: TFileStream;
  ScriptParser: TStreamTokenizer;
begin
  try
    S := TFileStream.Create(FScriptFileName, fmOpenRead);
    try
      ScriptParser := TStreamTokenizer.Create(S);
      try
        ScriptParser.OnLineProcessed := FOnLineProcessed;
        ParseScript(ScriptParser);
      finally
        ScriptParser.Free;
      end;
    finally
      S.Free;
    end;
  except
    on E: EFOpenError do
      ReportError(25, Self.Name, FScriptFileName, nil, True);
    on E: EAbort do
      ReportError(76, Self.Name, '', nil, True);
    else
      raise;
  end;
end;

constructor TForwardDef.Create(AName: String; AKind: TForwardKind);
begin
  inherited Create;
  Name := AName;
  Kind := AKind;
end;

procedure TCustomScriptRep.ParseScript(Parser: TStreamTokenizer);
var
  CurToken: TToken;
  _QueryCount: Integer;
  _ExprCount: Integer;
  _ControlCount: Integer;
  Forwards: TObjectList;

  procedure InitializeParser;
  var
    i: Integer;
  begin
    Forwards := TObjectList.Create;
    _QueryCount   := 0;
    _ExprCount    := 0;
    _ControlCount := 0;

    with Parser do
    begin
      AddStandAloneToken(StandAloneToken(#13#10, True));
      AddStandAloneToken(StandAloneToken(' ', True));
      AddStandAloneToken(StandAloneToken(#9, True));
      AddStandAloneToken(StandAloneToken(';', False));
      AddStandAloneToken(StandAloneToken('(', False));
      AddStandAloneToken(StandAloneToken(')', False));
      AddStandAloneToken(StandAloneToken('{', False));
      AddStandAloneToken(StandAloneToken('}', False));
      AddStandAloneToken(StandAloneToken(':=', False));
      AddStandAloneToken(StandAloneToken(',', False));
      AddStandAloneToken(StandAloneToken('&', False));
      AddStandAloneToken(SolidBlock('//', #13#10, True, False));
      AddStandAloneToken(SolidBlock('''', '''', False, True));
      AddStandAloneToken(SolidBlock('"', '"', False, True));
      AddStandAloneToken(SolidBlock('/*', '*/', True, False));

      FFuncList.Clear;

      if PlugInList <> nil then
        for i := 0 to PlugInList.Count-1 do
        begin
          with TExprDesc(PlugInList.Items[i]) do
            if AssumeSeparate then
              AddStandAloneToken(StandAloneToken(ReservedName, False));

          FFuncList.Add(TExprDesc(PlugInList.Items[i]));
        end;
    end;
  end;

  procedure FinalizeParser;
  begin
    if Forwards <> nil then Forwards.Free;
  end;

  function GetSysQueryName: String;
  begin
    Result := Format('SYS_QRY_%d', [_QueryCount]);
    Inc(_QueryCount);
  end;

  function GetSysExprName: String;
  begin
    Result := Format('SYS_EXPR_%d', [_ExprCount]);
    Inc(_ExprCount);
  end;

  function GetSysControlName: String;
  begin
    Result := Format('SYS_CTRL_%d', [_ControlCount]);
    Inc(_ControlCount);
  end;

  function ExtractString: String;
  begin
    with Parser do
    begin
      Next;
      Result := Token.Text;
      if Pos('"', Result) = 1 then
        Result := Copy(Result, 2, Length(Result)-2);
    end;
  end;

  function ExtractInteger: Integer;
  begin
    try
      Result := StrToInt(ExtractString);
    except
      with Parser.Token do
        ScriptError(29, Line, Col);
      raise;
    end;
  end;

  function FindSysExpression(AName: String;
    var DesiredArgKinds: TArgumentKinds): TRepExpressionClass;
  var
    i: Integer;
  begin
    Result := nil;

    with FFuncList do
      for i := 0 to Count-1 do
        with TExprDesc(Items[i]) do
          if (ReservedName = AnsiUpperCase(AName)) and
            (((ArgKinds - DesiredArgKinds) = []) or
            ((ArgKinds - DesiredArgKinds) = [akSet]))then
          begin
            DesiredArgKinds := ArgKinds;
            Result := ExprClass;
            break;
          end;
  end;
{
  function ValidateExprName(ExprName: String): Boolean;
  var
    i: Integer;
  begin
    Result := True;

    if RepExprName[ExprName] <> nil then exit;

    for i := 0 to Forwards.Count-1 do
      with TForwardDef(Forwards[i]) do
        if (Kind = fkExpression) and (Name = AnsiUpperCase(ExprName)) then exit;

    Result := False;
  end;
}
  function BuildExpr(Node: TRepExprNode): TRepToken;
  var
    ExprClass: TRepExpressionClass;
    RightToken: TRepToken;
    LeftToken:  TRepToken;
    ArgKinds: TArgumentKinds;
    CT: TConstType;
    SaveSeparator: Char;

    function ConstType(S: String): TConstType;
    begin
      if(S[1] in ['0'..'9']) then
        Result := ctDigit
      else if S[1] = '''' then
        Result := ctString
      else if (AnsiUpperCase(S) = 'TRUE') or (AnsiUpperCase(S) = 'FALSE') then
        Result := ctBool
      else
        result := ctBad;
    end;

  begin

    if Node = nil then
      with Parser.Token do ScriptError(31, Line, Col);

    ArgKinds := [];
    with Node do
    begin
      if Right <> nil then Include(ArgKinds, akRight);
      if Left <> nil then Include(ArgKinds, akLeft);
    end;

    ExprClass := FindSysExpression(Node.ExprName, ArgKinds);
    if (ExprClass = nil) and (ArgKinds <> []) then
      ScriptError(30, Node.Line, Node.Col);


    CT := ConstType(Node.ExprName);
    Result.Pos.Line := Node.Line;
    Result.Pos.Col  := Node.Col;

    if (ExprClass = nil) and (CT <> ctBad) then
    begin
      // 
      Result.Text := GetSysExprName;
      if CT = ctDigit then
      begin
        // 
        SaveSeparator := DecimalSeparator;
        DecimalSeparator := '.';
        try
          Result.Addr := AddExpression(TRepConstExpression.Create(Self,
                                                                  StrToFloat(Node.ExprName),
                                                                  Result.Text,
                                                                  RepPosition(Node.Line, Node.Col))
                                       );
        finally
          DecimalSeparator :=  SaveSeparator;
        end;
      end
      else if CT = ctBool then
        //  
        if AnsiUpperCase(Node.ExprName) = 'TRUE' then
          Result.Addr := AddExpression(TRepConstExpression.Create(Self,
                                                                  True,
                                                                  Result.Text,
                                                                  RepPosition(Node.Line,
                                                                              Node.Col)))
        else
          Result.Addr := AddExpression(TRepConstExpression.Create(Self,
                                                                  False,
                                                                  Result.Text,
                                                                  RepPosition(Node.Line, Node.Col)))
      else
        //  
        Result.Addr := AddExpression(TRepConstExpression.Create(Self,
                                                                Node.ExprName,
                                                                Result.Text,
                                                                RepPosition(Node.Line, Node.Col)));

      Result.AddrStr := AddrStr(Result);
    end
    else if ExprClass = nil then
    begin
      //    ,  -   
      // A + B   A  B   .
      Result.Text := Node.ExprName;
      Result.Addr := nil;
      Result.AddrStr := '0';
    end
    else if ExprClass <> nil then
    begin
      if Node.Right <> nil then
        RightToken := BuildExpr(Node.Right);

      if Node.Left <> nil then
        LeftToken := BuildExpr(Node.Left);

      Result.Text := GetSysExprName;

      if ([akLeft] = ArgKinds) or ([akRight] = ArgKinds) or ([] = ArgKinds) then
        //      (  ), 
        //   

        if akLeft in ArgKinds then
          //   - , : 10!
          Result.Addr := AddExpression(
                           TRepUnaryExpressionClass(ExprClass).Create(Self,
                                                                      LeftToken,
                                                                      EmptyRepToken, // WhenClause
                                                                      EmptyRepToken, // ResetClause
                                                                      Result.Text,
                                                                      RepPosition(Node.Line, Node.Col)
                                                                     )
                         )
        else if akRight in ArgKinds then
          //   - ,   
          Result.Addr := AddExpression(
                           TRepUnaryExpressionClass(ExprClass).Create(Self,
                                                                      RightToken,
                                                                      EmptyRepToken,
                                                                      EmptyRepToken,
                                                                      Result.Text,
                                                                      RepPosition(Node.Line, Node.Col))
                         )
        else
          //  ,  PageStart()
          Result.Addr := AddExpression(
                                       ExprClass.CreateOriginal(Self,
                                                                EmptyRepToken,
                                                                EmptyRepToken,
                                                                Result.Text,
                                                                RepPosition(Node.Line, Node.Col))
                         )
      else if ([akLeft, akRight] = ArgKinds) and (Node.ExprName <> '->') then
        //  ,  ->!
        if Node.ExprName = ',' then
        begin
          Result.Text := LeftToken.Text + ',' + RightToken.Text;
          Result.AddrStr := LeftToken.AddrStr + ',' + RightToken.AddrStr;
          Result.Addr := nil;
        end
        else
          Result.Addr := AddExpression(
                           TRepBinaryExpressionClass(ExprClass).Create(Self,
                                                                       LeftToken,
                                                                       RightToken,
                                                                       Result.Text,
                                                                       RepPosition(Node.Line, Node.Col))
                         )
      else if ([akLeft, akRight] = ArgKinds) and (Node.ExprName = '->') then
      begin
        //  ->
        Result.Addr := AddExpression(
                          TRepFieldExpressionClass(ExprClass).Create(Self,
                                                                     RepPosition(Node.Line, Node.Col),
                                                                     LeftToken,
                                                                     RightToken)
                       );
        Result.Text := Result.Addr.Name;
      end
      else
        //    
        Result.Addr := AddExpression(
                         TRepSetExpressionClass(ExprClass).Create(Self,
                                                                  RightToken,
                                                                  Result.Text,
                                                                  RepPosition(Node.Line, Node.Col))
                       );
      if Node.ExprName <> ',' then
        Result.AddrStr := AddrStr(Result);
    end;

  end;

  function ExtractExpression(ExprBreak: String): TRepToken;
  begin
    Parser.Next;
    with TRepExprParser.Create(Parser, Self) do
      try
        BuildTree(ExprBreak);
        Result := BuildExpr(RootNode);
      finally
        Free;
      end;
  end;

  function ExtractStringList(StringList: TStrings; ProcessAmpersand: Boolean): TRepPosition;
  var
    S: String;
    IgnoreNext: Boolean;
    ExprToken: TRepToken;
  begin
    StringList.Clear;

    S := EmptyStr;
    IgnoreNext := False;

    with Parser, StringList do
    begin
      with Token do
      begin
        if Text <> '{' then ScriptError(32, Line, Col);
        Result.Line := Line;
        Result.Col  := Col;
      end;

      EnableHidden := True;
      DisableStandAloneToken(SolidBlock('"', '"', False, True));
      DisableStandAloneToken(SolidBlock('''', '''', False, True));
      DisableStandAloneToken(SolidBlock('//', #13#10, False, True));
      DisableStandAloneToken(SolidBlock('/*', '*/', False, True));

      Next;
      S := Token.Text;
      if Pos(#13+#10, S) = 1 then
      begin
        S := Copy(S, 3, Length(S));
        Inc(Result.Line);
      end;

      if S <> '&' then
        Next
      else
        S := '';

      while True do
      begin
        if EOS then
          with Token do
            ScriptError(33, Line, Col);

        if Token.Text = '}' then
        begin
          Add(S);
          break;
        end;

        if ProcessAmpersand then
          if Token.Text = '&' then
            if not IgnoreNext then
            begin
              EnableStandAloneToken(SolidBlock('"', '"', False, True));
              EnableStandAloneToken(SolidBlock('''', '''', False, True));
              EnableStandAloneToken(SolidBlock('//', #13#10, False, True));
              EnableStandAloneToken(SolidBlock('/*', '*/', False, True));
              EnableHidden := False;

              ExprToken := ExtractExpression('& PICTURE');

              EnableHidden := True;
              DisableStandAloneToken(SolidBlock('"', '"', False, True));
              DisableStandAloneToken(SolidBlock('''', '''', False, True));
              DisableStandAloneToken(SolidBlock('//', #13#10, False, True));
              DisableStandAloneToken(SolidBlock('/*', '*/', False, True));

              S := S + '&'+ExprToken.Text+','+ExprToken.AddrStr+','+IntToStr(ExprToken.Pos.Line)+','+IntToStr(ExprToken.Pos.Col)+' ';
              IgnoreNext := True;
              continue;
            end
            else
              IgnoreNext := False;

        S := S + Token.Text;
        Next;
      end;
      EnableStandAloneToken(SolidBlock('"', '"', False, True));
      EnableStandAloneToken(SolidBlock('''', '''', False, True));
      EnableStandAloneToken(SolidBlock('//', #13#10, False, True));
      EnableStandAloneToken(SolidBlock('/*', '*/', False, True));
      EnableHidden := False;
    end;
  end;

  {
    Process report options

    REPORT [SPOOL <spool_file_name> [APPEND]] [PAGESIZE <page_size> [NOALIGN]]
    [TITLE <report_title>]
  }

  procedure ParseRepDefinition;
  var
    AToken: TToken;
    Text: String;
  begin
    FTitle := EmptyStr;

    while True do
    begin
      if Parser.EOS then break;

      AToken := Parser.Token;
      Text := AnsiUpperCase(AToken.Text);

      if Text = 'TITLE' then
      begin
        FTitle := ExtractString;
        Parser.Next;
      end
      else if Text = 'PAGESIZE' then
      begin
        PageHeight := ExtractInteger;
        Parser.Next;
        if AnsiUpperCase(Parser.Token.Text) = 'NOALIGN' then
        begin
          AlignPages := False;
          Parser.Next;
        end;
      end
      else if Text = 'SPOOL' then
      begin
        FOutputFileName := ExtractString;
        Parser.Next;
        if AnsiUpperCase(Parser.Token.Text) = 'APPEND' then
        begin
          FAppendFile := True;
          Parser.Next;
        end
      end
      else if Text = ';' then
        break
      else
        with AToken do ScriptError(34, Line, Col);
    end;

    if PageHeight = 0 then
    begin
      with Parser.Token do ScriptError(34, Line, Col);
    end
    else if FOutputFileName = EmptyStr then
    begin
      with Parser.Token do ScriptError(35, Line, Col);
    end;
  end;

(*
   this routine processes query definition from script stream
   QUERY [<query_name>] [FETCH WHEN <expr_name>] [FETCH BY <rowset_size> ROWS]
   [RESET WHEN <expr_name>] [[UPDATES PARAMS] [ALIAS "<alias_name>"]|
   [REF "<data_set_name>"]]
   [CURSOR "<cursor_name>"]
   {<query_body>};
*)

  procedure ProcessQueryDef;
  var
    S: String;
    QueryName: String;
    QueryText: String;
    FetchClauseToken: TRepToken;
    ResetClauseToken: TRepToken;
    ModifyPar: Boolean;
    IsBDE    : Boolean;
    AliasName: String;
    HasRef: Boolean;
    RefDataSetName: String;
    Str: TStringList;
    RowsetSize: Integer;
    CursorName: String;
  begin
    QueryName := EmptyStr;
    QueryText := EmptyStr;
    FetchClauseToken := EmptyRepToken;
    ResetClauseToken := EmptyRepToken;
    ModifyPar := False;
    IsBDE     := False;
    AliasName := EmptyStr;
    RefDataSetName := EmptyStr;
    CursorName := EmptyStr;
    HasRef := False;
    RowSetSize := -1;

    Str := TStringList.Create;

    Parser.Next;

    while True do
    begin

      if Parser.EOS then
        with Parser.Token do ScriptError(36, Line, Col);

//      SaveCurIndex := CurTokenIndex;
      S := Parser.Token.Text;

      if AnsiUpperCase(S) = '{' then break;

      if AnsiUpperCase(S) = 'UPDATES' then
      begin
        if ModifyPar then
          with Parser.Token do ScriptError(37, Line, Col);

        Parser.Next;

        if AnsiUpperCase(Parser.Token.Text) <> 'PARAMS' then
          with Parser.Token do ScriptError(38, Line, Col);

        ModifyPar := True;

        Parser.Next;
      end
      else if AnsiUpperCase(S) = 'FETCH' then
      begin
        if ResetClauseToken.Text <> EmptyStr then
          with Parser.Token do ScriptError(36, Line, Col);

        Parser.Next;

        with Parser.Token do
          if AnsiUpperCase(Text) = 'WHEN' then
            FetchClauseToken := ExtractExpression('RESET { UPDATES // ALIAS REF FETCH CURSOR')
          else if AnsiUpperCase(Text) = 'BY' then
          begin
            RowSetSize := ExtractInteger;
            Parser.Next;
          end
          else
            ScriptError(65, Line, Col);
      end
      else if AnsiUpperCase(S) = 'RESET' then
      begin
        Parser.Next;

        with Parser.Token do
          if AnsiUpperCase(Text) <> 'WHEN' then ScriptError(39, Line, Col);

        ResetClauseToken := ExtractExpression('{ FETCH UPDATES // ALIAS REF CURSOR');
      end
      else if AnsiUpperCase(S) = 'ALIAS' then
      begin
        AliasName := ExtractString;
        IsBDE     := True;
        Parser.Next;
      end
      else if AnsiUpperCase(S) = 'REF' then
      begin
        RefDataSetName := ExtractString;
        HasRef         := True;
        Parser.Next;
      end
      else if AnsiUpperCase(S) = 'CURSOR' then
      begin
        if CursorName <> '' then
          with Parser.Token do ScriptError(69, Line, Col);

        CursorName := ExtractString;
        Parser.Next;
      end
      else
      begin
        if (QueryName <> EmptyStr) or
           (FetchClauseToken.Text <> EmptyStr) or
           (ResetClauseToken.Text <> EmptyStr) then
          with Parser.Token do ScriptError(40, Line, Col);
        QueryName := S;
        Parser.Next;
      end;
    end;

    if QueryName = EmptyStr then QueryName := GetSysQueryName;

    with AddQuery(TRepQuery.CreateOriginal(Self,
                                           FetchClauseToken,
                                           ResetClauseToken,
                                           QueryName,
                                           CursorName)) do
    begin
      HasReference := HasRef;
      RefName := RefDataSetName;
      SQLPosition := ExtractStringList(Str, True);
      ModifyParams := ModifyPar;
      UnderBDE := IsBDE;
      Alias := AliasName;
      FetchBy := RowsetSize;
      with Parser.Token do
        if (FetchBy <> RowsetSize) and (RowSetSize <> -1) then
          ScriptError(66, Line, Col);
      SQL := Str;
    end;

    Str.Free;

    with Parser do
    begin
      if EOS then
        with Token do ScriptError(36, Line, Col);

      Next;

      with Token do
        if Text <> ';' then ScriptError(41, Line, Col);
    end;
  end;

{*
    CONTROL [<control_name>] [ALWAYS ON PAGE] [FINISH PAGE [DISABLE FOOTER]]
            [TURN WHEN <expr_name>] [ORDER HIGHEST | HIGH | NORMAL | LOW | DEFAULT]
    begin
      <control_body>
    end;
*}
  procedure ProcessControlDef;
  var
    CtrlName: String;
    WhenClauseToken: TRepToken;
    OnPage: Boolean;
    FinPage: Boolean;
    DisableFtr: Boolean;
    S: string;
    SL: TStringList;
    ControlOrder: TControlOrder;
  begin
    CtrlName := EmptyStr;
    WhenClauseToken := EmptyRepToken;
    OnPage := False;
    FinPage := False;
    DisableFtr := False;
    ControlOrder := coDefault;

    Parser.Next;

    with Parser do
      while True do
      begin
        if EOS then
          with Token do ScriptError(42, Line, Col);

        S := Token.Text;

        if AnsiUpperCase(S) = '{' then break;

        if AnsiUpperCase(S) = 'DISABLE' then
        begin
          if CtrlName = EmptyStr then
            with Token do ScriptError(43, Line, Col);

          if not FinPage then
            with Token do ScriptError(44, Line, Col);

          Next;
          with Token do
            if AnsiUpperCase(Text) <> 'FOOTER' then ScriptError(45, Line, Col);

          if DisableFtr then
            with Token do ScriptError(46, Line, Col);
          DisableFtr := True;
          Next;
        end
        else if AnsiUpperCase(S) = 'FINISH' then
        begin
          if CtrlName = EmptyStr then
            with Token do ScriptError(43, Line, Col);

          Next;
          with Token do
            if AnsiUpperCase(Text) <> 'PAGE' then ScriptError(44, Line, Col);

          if FinPage then
            with Token do ScriptError(46, Line, Col);

          FinPage := True;
          Next;
        end
        else if AnsiUpperCase(S) = 'ALWAYS' then
        begin
          if CtrlName = EmptyStr then
            with Token do ScriptError(43, Line, Col);

          Next;
          with Token do
            if AnsiUpperCase(Text) <> 'ON' then ScriptError(47, Line, Col);

          Next;
          with Token do
            if AnsiUpperCase(Text) <> 'PAGE' then ScriptError(47, Line, Col);
          OnPage := True;
          Next;
        end
        else if AnsiUpperCase(S) = 'TURN' then
        begin
          if CtrlName = EmptyStr then
            with Token do ScriptError(43, Line, Col);

          Next;
          with Token do
            if AnsiUpperCase(Text) <> 'WHEN' then ScriptError(48, Line, Col);

          WhenClauseToken := ExtractExpression('{ ALWAYS FINISH ORDER //');
        end
        else if AnsiUpperCase(S) = 'ORDER' then
        begin
          Next;
          S := AnsiUpperCase(Token.Text);
          if S = 'HIGHEST' then
            ControlOrder := coHighest
          else if S = 'HIGH' then
            ControlOrder := coHigh
          else if S = 'NORMAL' then
            ControlOrder := coNormal
          else if S = 'LOW' then
            ControlOrder := coLow
          else if S = 'DEFAULT' then
            ControlOrder := coDefault
          else
            with Token do ScriptError(49, Line, Col);
          Next;
        end
        else if CtrlName = '' then
        begin
          CtrlName := S;
          Next;
        end
        else
          with Token do ScriptError(40, Line, Col);
      end;

      if CtrlName = EmptyStr then
        with Parser.Token do ScriptError(43, Line, Col);

      SL := TStringList.Create;
      try
        ExtractStringList(SL, True);
        with AddControl(TExprRepControl.Create(Self,
                                               WhenClauseToken,
                                               CtrlName, SL)) do
        begin
          if ControlOrder = coDefault then
            if OnPage then
              Order := NLowControlRange + Self.RepControlCount
            else if WhenClauseToken.Text <> EmptyStr then
              Order := NMidControlRange + Self.RepControlCount
            else
              Order := NHighControlRange + Self.RepControlCount
          else
            case ControlOrder of
              coHighest:
                Order := NLowestControlRange + Self.RepControlCount;
              coHigh:
                Order := NLowControlRange + Self.RepControlCount;
              coNormal:
                Order := NMidControlRange + Self.RepControlCount;
              coLow:
                Order := NHighControlRange + Self.RepControlCount;
            end;

          AlwaysOnPage := OnPage;
          FinishesPage := FinPage;
          DisableFooter := DisableFtr;
        end;
      finally
        SL.Free;
      end;

    with Parser do
    begin
      if EOS then
        with Token do ScriptError(42, Line, Col);

      Next;

      with Token do
        if Text <> ';' then ScriptError(41, Line, Col);
    end;
  end;
  {
    PARAM  param_name_1 [mask "<mask>"][, param_def_2];
  }
  procedure ProcessParamName;
  var
    S: String;
    AParamName: String;
    AnInputMask: String;
    Pos: TRepPosition;
  begin

    AParamName := EmptyStr;
    anInputMask := EmptyStr;

    while True do
      with Parser do
      begin
        if EOS then
          with Token do ScriptError(52, Line, Col);

        with Token do Pos := RepPosition(Line, Col);

        S := AnsiUpperCase(ExtractString);

        if (S = ';') or (S = ',') then
        begin
          if AParamName = EmptyStr then
            with Token do ScriptError(53, Line, Col);

          if AddParamToList then
            with Params.CreateParam(AParamName, Pos) do
              InputMask := AnInputMask;

          AParamName := EmptyStr;
          AnInputMask := EmptyStr;

          if S = ';' then
            break;

        end
        else if S = 'MASK' then
        begin
          if AParamName = EmptyStr then
            with Token do ScriptError(53, Line, Col);

          AnInputMask := ExtractString;
        end
        else
          AParamName := S;
      end;

  end;

  {
    <expr_name> := <expr_body> [WHEN|BEFORE|AFTER <expr_name> RESET TO <expr_name>] ;
  }
  procedure ProcessExprDef;
  var
    AnExprName: String;
    ChildExprToken: TRepToken;
    WhenClauseToken: TRepToken;
    ResetClauseToken: TRepToken;
    S: String;
    Before: Boolean;
    After: Boolean;
    Pos: TRepPosition;
  begin
    AnExprName := Parser.Token.Text;

    ChildExprToken   := EmptyRepToken;
    WhenClauseToken  := EmptyRepToken;
    ResetClauseToken := EmptyRepToken;
    Before := False;
    After := False;

    Parser.Next;

    while True do
      with Parser do
      begin
        if EOS then
          with Token do ScriptError(54, Line, Col);

        S := Token.Text;
        if S = ':=' then
        begin
          if ChildExprToken.Text <> EmptyStr then
            with Token do ScriptError(40, Line, Col);

          ChildExprToken := ExtractExpression('BEFORE AFTER WHEN RESET ;');
        end
        else if (AnsiUpperCase(S) = 'WHEN') or
                (AnsiUpperCase(S) = 'BEFORE') or
                (AnsiUpperCase(S) = 'AFTER') then
        begin
          if ChildExprToken.Text = EmptyStr then
            with Token do ScriptError(55, Line, Col);

          WhenClauseToken := ExtractExpression('RESET ;');
          Before := (AnsiUpperCase(S) = 'BEFORE');
          After  := (AnsiUpperCase(S) = 'AFTER');
        end
        else if AnsiUpperCase(S) = 'RESET' then
        begin
          if ChildExprToken.Text = EmptyStr then
            with Token do ScriptError(55, Line, Col);

          Next;

          if EOS then
            with Token do ScriptError(54, Line, Col);

          with Token do
            if AnsiUpperCase(Text) <> 'TO' then ScriptError(56, Line, Col);

          ResetClauseToken := ExtractExpression('WHEN BEFORE AFTER ;');
        end
        else if S = ';' then
          break
        else
          with Token do
            ScriptError(40, Line, Col);
      end;

    if ChildExprToken.Text = EmptyStr then
      with Parser.Token do ScriptError(55, Line, Col);

    with Parser.Token do Pos := RepPosition(Line, Col);

    with AddExpression(TRepDuplicateExpr.Create(Self,
                                                ChildExprToken,
                                                WhenClauseToken,
                                                ResetClauseToken,
                                                AnExprName,
                                                Pos)) do
    begin
      ResetBefore := Before;
      ResetAfter  := After;
    end;

  end;

  {
    SUBREPORT <name> AT FILE <file_path> [USE PARENT OUTPUT] [RUN WHEN <expr_name>]
     [NOWAIT];
  }
  procedure ProcessSubreport;
  var
    SubRepName: String;
    FileName: String;
    AParentOutput: Boolean;
    SubRepWhenToken: TRepToken;
    S: String;
    RepNoWait: Boolean;
    NameLine, NameCol: Integer;
  begin
    SubRepName := EmptyStr;
    FileName   := EmptyStr;
    AParentOutput := False;
    SubRepWhenToken := EmptyRepToken;
    RepNoWait := False;
    NameLine := -1;
    NameCol  := -1;

    Parser.Next;

    with Parser do
      while True do
      begin
        S := AnsiUpperCase(Token.Text);

        if S = ';' then break;

        if S = 'AT' then
        begin
          if FileName <> EmptyStr then
            with Token do ScriptError(40, Line, Col);

          if SubRepName = EmptyStr then
            with Token do ScriptError(57, Line, Col);

          Next;

          with Token do
            if AnsiUpperCase(Text) <> 'FILE' then ScriptError(58, Line, Col);

          FileName := ExtractString;
          Next;
        end
        else if S = 'USE' then
        begin

          if AParentOutput then
            with Token do ScriptError(40, Line, Col);

          if SubRepName = EmptyStr then
            with Token do ScriptError(57, Line, Col);

          Next;

          with Token do
            if AnsiUpperCase(Text) <> 'PARENT' then ScriptError(59, Line, Col);

          Next;

          with Token do
            if AnsiUpperCase(Text) <> 'OUTPUT' then ScriptError(59, Line, Col);
          AParentOutput := True;

          Next;
        end
        else if S = 'RUN' then
        begin
          if SubRepWhenToken.Text <> EmptyStr then
            with Token do ScriptError(40, Line, Col);

          if SubRepName = EmptyStr then
            with Token do ScriptError(57, Line, Col);

          Next;

          with Token do
            if AnsiUpperCase(Text) <> 'WHEN' then ScriptError(60, Line, Col);

          SubRepWhenToken := ExtractExpression('AT USE NOWAIT ;');
        end
        else if S = 'NOWAIT' then
        begin
          RepNoWait := True;
          Next;
        end
        else
          if SubRepName <> EmptyStr then
            with Token do ScriptError(40, Line, Col)
          else
          begin
            SubRepName := S;
            with Token do
            begin
              NameLine := Line;
              NameCol  := Col;
            end;
            Next;
          end;
      end;

      if SubRepName = EmptyStr then
        with Parser.Token do ScriptError(57, Line, Col);

      if FileName = EmptyStr then
        with Parser.Token do ScriptError(61, Line, Col);

      with TCustomScriptRep(
        AddSubreport(TCustomScriptRep.CreateSubreport(Self, Self))
        ) do
      begin
        WhenClause := SubRepWhenToken;
        try
          Name := SubRepName;
        except
          ScriptError(75, NameLine, NameCol);
        end;
        ScriptFile := FileName;
        OracleSession := Self.OracleSession;
        ParentOutput := AParentOutput;
        NoWait := RepNoWait;
      end
  end;

begin
  InitializeParser;

  with Parser do
    try
      if EOS or (AnsiUpperCase(Token.Text) <> 'REPORT') then
        ScriptError(62, -1, -1)
      else
      begin
        Next;
        ParseRepDefinition;
        Next;
      end;


      while not EOS do
      begin
        CurToken := Token;
        if AnsiUpperCase(CurToken.Text) = 'QUERY' then
          ProcessQueryDef
        else if AnsiUpperCase(CurToken.Text) = 'CONTROL' then
          ProcessControlDef
        else if AnsiUpperCase(CurToken.Text) = 'PARAM' then
          ProcessParamName
        else if AnsiUpperCase(CurToken.Text) = 'SUBREPORT' then
          ProcessSubreport
        else
          ProcessExprDef;

        Next;
      end;
    finally
      FinalizeParser;
    end;

end;

procedure TCustomScriptRep.SetPrepared(Value: Boolean);
begin
  if FScriptFileName = EmptyStr then
    if not (csReading in ComponentState) then
      raise EScripFileEmpty.Create('Specify script file name, please')
    else
      exit;

  try
    inherited SetPrepared(Value);
  except
    on EReportTerminated do
      if (csDesigning in ComponentState) then
        OpenErrors;
    else
      raise;
  end;
end;

procedure TCustomScriptRep.OpenErrors;
var
  F: TFmRepErrors;
begin
  if HasErrors then
  begin
    F := TfmRepErrors.Create(nil);
    if F.sgErrors = nil then
      MessageDlg('F.gsErrors = nil!!!!!!!', mtError, [mbCancel], 0);

    F.ShowErrors(ErrorList);

    F.Free;
  end;
end;

procedure TCustomScriptRep.OpenCompiler;
begin
  fmDebugger := TfmDebugger.Create(nil);
  try
    fmDebugger.OpenDebugger(ScriptFile, OracleSession);
  finally
    fmDebugger.Free;
  end;
end;

procedure TCustomScriptRep.DoBeforeSubRepParsed(Index: Integer);
begin
  if Assigned(FBeforeSubReportParsed) then
    FBeforeSubReportParsed(Self, TCustomScriptRep(ChildRep[Index]));
end;

procedure TCustomScriptRep.ScriptError(ErrorCode: Integer; Line, Col: Integer);
var
  E: TRepExpression;
begin
  E := TRepExpression.CreateOriginal(Self, EmptyRepToken, EmptyRepToken, 'XXXXX', RepPosition(Line, Col));
  try
    ReportError(ErrorCode, Self.Name, '', E, True);
  finally
    E.Free;
  end;
end;

procedure AddPlugIn(FuncName: String;
                    Args: TArgumentKinds;
                    ExprClass: TRepExpressionClass;
                    IsSeparate: Boolean);

  procedure FindTheSameName(AName: String);
  var
    i: Integer;
  begin
    for i := 0 to PlugInList.Count-1 do
      with TExprDesc(PlugInList.Items[i]) do
        if AName = ReservedName then
          raise EDuplicatePlugInName.Create('Duplicate PlugIn name');
  end;
var
  Desc: TExprDesc;
begin
//  FindTheSameName(FuncName);
  InitPlugInList;
  Desc := TExprDesc.Create(FuncName, Args, ExprClass);
  Desc.AssumeSeparate := IsSeparate;
  PlugInList.Add(Desc);
end;

initialization
  InitPlugInList;

finalization
  PlugInList.Free;
end.

