{**
*
*         Script Reporter Component library
*         Stream tokenizer class
*          Vlad Ilushchenko 1998-99
*
*
**}

unit Tokenizer;

interface

uses Classes;

type
  TTokenKind = (tkToken, tkSymbol, tkSolidBlock, tkError);

  TStreamTokenizer = class;
  TLineProcessedEvent = procedure(Sender: TObject; Line: Integer) of Object;

  TToken = class
  public
    Text: String;
    Line: Integer;
    Col:  Integer;
    StreamPos: Integer;
    Kind: TTokenKind;
  private
    StopLine: Integer;
    StopCol:  Integer;
    InternalLen: Integer;
  end;

  TStandAloneToken = class
  private
    Text: String;
    TermText: String;
    Hidden: Boolean;
    DecodePaired: Boolean;
    Enabled: Boolean;
  end;

  TTokenScanResult = (srMany, srExact, srNone);
  TTokenSearchResult = record
    MaxLen: Integer;
    TopPos: Integer;
    ScanResult: TTokenScanResult;
  end;

  TStreamTokenizer = class
  private
    FStream: TStream;
    FStandAloneTokens: TList;
    FToken: TToken;
    FCreating: Boolean;
    FStreamSize: Integer;
    FEnableHidden: Boolean;
    FOnLineProcessed: TLineProcessedEvent;
  protected
    procedure GetNextToken;
    procedure GetPrevToken;
    procedure SetStream(Value: TStream);

    function FindStandAloneToken(Ch: Char): TTokenSearchResult;
    function ScanForStandAloneToken(P: PChar; Pos: Integer): Integer;

    function TokenByToken(StandAloneToken: TStandAloneToken): Integer;
    procedure DoLineProcessed(Line: Integer);
  public
    constructor Create(AStream: TStream);
    destructor Destroy; override;

    procedure AddStandAloneToken(StandAloneToken: TStandAloneToken);
    procedure DisableStandAloneToken(StandAloneToken: TStandAloneToken);
    procedure EnableStandAloneToken(StandAloneToken: TStandAloneToken);

    procedure Reset;

    function BOS: Boolean;
    function EOS: Boolean;
    function Token: TToken;

    procedure Next;
    procedure Prev;
    procedure First;
    procedure Last;

    procedure Dump(const FileName: String);

    property Stream: TStream read FStream write SetStream;
    property EnableHidden: Boolean read FEnableHidden write FEnableHidden;
    property OnLineProcessed: TLineProcessedEvent read FOnLineProcessed
                                                  write FOnLineProcessed;
  end;

function StandAloneToken(Text: String; Hidden: Boolean): TStandAloneToken;
function SolidBlock(StartText, EndText: String; Hidden: Boolean; DecodePaired: Boolean): TStandAloneToken;

implementation

uses SysUtils, ObjList;

function StandAloneToken(Text: String; Hidden: Boolean): TStandAloneToken;
begin
  Result              := TStandAloneToken.Create;
  Result.Text         := Text;
  Result.TermText     := '';
  Result.Hidden       := Hidden;
  Result.DecodePaired := False;
  Result.Enabled      := True;
end;

function SolidBlock(StartText, EndText: String; Hidden: Boolean; DecodePaired: Boolean): TStandAloneToken;
begin
  Result              := TStandAloneToken.Create;
  Result.Text         := StartText;
  Result.TermText     := EndText;
  Result.Hidden       := Hidden;
  Result.DecodePaired := DecodePaired;
  Result.Enabled      := True;
end;

function CompareStandAloneTokens(Item1, Item2: Pointer): Integer;
begin
  Result := StrComp(PChar(TStandAloneToken(Item2).Text),
                     PChar(TStandAloneToken(Item1).Text));
  if Result = 0 then
    Result := StrComp(PChar(TStandAloneToken(Item2).termText),
                     PChar(TStandAloneToken(Item1).termText));
end;

constructor TStreamTokenizer.Create(AStream: TStream);
begin
  inherited Create;
  FStandAloneTokens := TList.Create;
  FToken            := TToken.Create;
  FEnableHidden     := False;
  FCreating := True;
  try
    SetStream (AStream);
  finally
    FCreating := False;
  end;
  FToken.Text := '';
  FToken.Kind := tkError;
end;

destructor TStreamTokenizer.Destroy;
begin
  Reset;
  FStandAloneTokens.Free;
  FToken.Free;
  SetStream(nil);
  inherited Destroy;
end;

procedure TStreamTokenizer.SetStream(Value: TStream);
begin
  FStream := Value;
  if FStream <> nil then
  begin
    FStream.Position := 0;
    FStreamSize := FStream.Size;
    if not FCreating then GetNextToken;
  end;
end;

procedure TStreamTokenizer.AddStandAloneToken(StandAloneToken: TStandAloneToken);
begin
  FStandAloneTokens.Add(StandAloneToken);
  FStandAloneTokens.Sort(CompareStandAloneTokens);
end;

procedure TStreamTokenizer.DoLineProcessed(Line: Integer);
begin
  if Assigned(FOnLineProcessed) then FOnLineProcessed(Self, Line);
end;

function TStreamTokenizer.TokenByToken(StandAloneToken: TStandAloneToken): Integer;
var
  i: Integer;
begin
  Result := -1;
  for i := 0 to FStandAloneTokens.Count-1 do
    with TStandAloneToken(FStandAloneTokens.Items[i]) do
      if (Text = StandAloneToken.Text) and
         (TermText = StandAloneToken.TermText) then
      begin
        Result := i;
        break;
      end;
end;                  

procedure TStreamTokenizer.DisableStandAloneToken(StandAloneToken: TStandaloneToken);
var
  i: Integer;
begin
  try
    i := TokenByToken(StandAloneToken);
    if i > -1 then
      TStandAloneToken(FStandAloneTokens.Items[i]).Enabled := False;
  finally
    StandAloneToken.Free;
  end;
end;

procedure TStreamTokenizer.EnableStandAloneToken(StandAloneToken: TStandaloneToken);
var
  i: Integer;
begin
  try
    i := TokenByToken(StandAloneToken);
    if i > -1 then
      TStandAloneToken(FStandAloneTokens.Items[i]).Enabled := True;
  finally
    StandAloneToken.Free;
  end;
end;

procedure TStreamTokenizer.Reset;
var
  i: Integer;
  P: Pointer;
begin
  for i := 0 to FStandAloneTokens.Count-1 do
  begin
    P := FStandAloneTokens.Items[i];
    if P <> nil then
    begin
      TStandAloneToken(P).Free;
      FStandAloneTokens.Items[i] := nil;
    end;
  end;

  FStandAloneTokens.Clear;

  FStream := nil;
  FToken.Kind := tkError;
  FToken.Text := '';
  FToken.Line := 0;
  FToken.Col  := 0;
  FToken.StopLine := 0;
  FToken.StopCol  := 0;
end;

function TStreamTokenizer.BOS: Boolean;
begin
  if FStream = nil then
    Result := True
  else
    Result := FStream.Position = 0;
end;

function TStreamTokenizer.EOS: Boolean;
begin
  if FStream = nil then
    Result := True
  else
    Result := (FStream.Position >= FStreamSize) and (FToken.Kind = tkError);
end;

function TStreamTokenizer.Token: TToken;
begin
  if FToken.Kind = tkError then Next;
  Result := FToken;
end;

procedure TStreamTokenizer.Next;
begin
  if FStream <> nil then
    GetNextToken;
end;

procedure TStreamTokenizer.Prev;
begin
  if FStream <> nil then
    GetPrevToken;
end;

procedure TStreamTokenizer.First;
begin
  if FStream <> nil then
  begin
    FStream.Position := 0;
    GetNextToken;
  end;
end;

procedure TStreamTokenizer.Last;
begin
  if FStream <> nil then
  begin
    FStream.Position := FStreamSize;
    GetPrevToken;
  end;
end;

function TStreamTokenizer.FindStandAloneToken(Ch: Char): TTokenSearchResult;
  function FindFirst(StartPos, EndPos: Integer; Ch: Char): Integer;
  begin
    Result := StartPos + (EndPos - StartPos) div 2;

    if StartPos > EndPos then
      Result := -1
    else if StartPos = EndPos - 1 then
    begin
      if (TStandAloneToken(FStandAloneTokens[StartPos]).Text[1] = Ch) and
         TStandAloneToken(FStandAloneTokens[StartPos]).Enabled then
        Result := StartPos
      else if (TStandAloneToken(FStandAloneTokens[EndPos]).Text[1] = Ch) and
              TStandAloneToken(FStandAloneTokens[EndPos]).Enabled then
        Result := EndPos
      else
        Result := -1;
    end
    else
      with TStandAloneToken(FStandAloneTokens[Result]) do
        if (Text[1] < Ch) and (StartPos < EndPos) then
          Result := FindFirst(StartPos, Result, Ch)
        else if (Text[1] > Ch) and (StartPos < EndPos) then
          Result := FindFirst(Result, EndPos, Ch)
        else if (Text[1] = Ch) and Enabled then
          exit
        else
          Result := -1;
  end;
var
  l, i: Integer;
begin
  l := FindFirst(0, FStandAloneTokens.Count-1, Ch);
  if l = -1 then
    Result.ScanResult := srNone
  else
  begin
    i := l;
    Result.MaxLen := Length(TStandAloneToken(FStandAloneTokens[l]).Text);
    Result.TopPos := l-1;
    if TStandAloneToken(FStandAloneTokens[l]).Text = Ch then
      Result.ScanResult := srExact
    else
      Result.ScanResult := srMany;

    while (i >= 0) and (TStandAloneToken(FStandAloneTokens[i]).Text[1] = Ch) do
    begin
      Result.TopPos := i;
      Result.MaxLen := Max(Result.MaxLen, Length(TStandAloneToken(FStandAloneTokens[i]).Text));
      Dec(i);
    end;

    if Result.TopPos < l then
      Result.ScanResult := srMany;

    i := l+1;
    while (i < FStandAloneTokens.Count) and
          (TStandAloneToken(FStandAloneTokens[i]).Text[1] = Ch) do
    begin
      Result.MaxLen := Max(Result.MaxLen, Length(TStandAloneToken(FStandAloneTokens[i]).Text));
      Inc(i);
    end;

    if i > l + 1 then
      Result.ScanResult := srMany;
  end;
end;

function TStreamTokenizer.ScanForStandAloneToken(P: PChar; Pos: Integer): Integer;
begin
  Result := -1;
  while (Pos < FStandAloneTokens.Count) and
        (TStandAloneToken(FStandAloneTokens[Pos]).Text[1] = P^) do
    with TStandAloneToken(FStandAloneTokens[Pos]) do
      if (StrLIComp(P, PChar(Text), Length(Text)) = 0) and Enabled then
      begin
        Result := Pos;
        break;
      end
      else
        Inc(Pos);
end;

procedure TStreamTokenizer.GetNextToken;
var
  S: array [0..255] of char;
  P: PChar;
  SearchResult: TTokenSearchResult;
  SavePos: Integer;
  Pos: Integer;
  PairText: String;

  function CompareTo(CompareText: String): Boolean;
  begin
    Result := False;

    if CompareText[1] = P^ then
    begin
      SavePos := FStream.Position;
      FStream.ReadBuffer((P+1)^, Min(Length(CompareText)-1, FStreamSize - FStream.Position));
      if StrLIComp(P, PChar(CompareText), Length(CompareText)) = 0 then
        Result := True
      else
        FStream.Position := SavePos;
    end;
  end;

begin
  P := @S;
  S[1] := #0;

  Stream.Position := FToken.StreamPos + FToken.InternalLen;
  FToken.Col      := FToken.StopCol;
  FToken.Line     := FToken.StopLine;

  FToken.Text := '';
  FToken.StreamPos := Stream.Position;
  FToken.Kind := tkError;
  FToken.InternalLen := 0;

  while FStream.Position < FStreamSize do
  begin
    FStream.ReadBuffer(P^, 1);
    P[1] := #0;
    SearchResult := FindStandAloneToken(S[0]);

    if SearchResult.ScanResult = srMany then
    begin
      SavePos := FStream.Position;
      FStream.ReadBuffer((P+1)^,
                         Min(Length(TStandAloneToken(FStandAloneTokens[SearchResult.TopPos]).Text)-1,
                             FStreamSize - FStream.Position));
      Pos := ScanForStandAloneToken(P, SearchResult.TopPos);
      if Pos = -1 then
      begin
        P[1] := #0;
        FStream.Position := SavePos;
      end
      else
        FStream.Position := SavePos + Length(TStandAloneToken(FStandAloneTokens[Pos]).Text)-1;
      SearchResult.TopPos := Pos;
    end;

    if (SearchResult.ScanResult = srExact) or
       ((SearchResult.ScanResult = srMany) and (SearchResult.TopPos <> -1)) then
      if FToken.Text <> '' then
      begin
        FToken.Kind := tkToken;
        FToken.InternalLen := Length(FToken.Text);
        break;
      end
      else
        with TStandAloneToken(FStandAloneTokens[SearchResult.TopPos]) do
        begin
          if Text = #13#10 then
          begin
            Inc(FToken.StopLine);
            DoLineProcessed(FToken.StopLine);
            FToken.StopCol := -2;
          end;

          if TermText <> '' then
          begin
            FToken.Text := Text;
            FToken.Kind := tkSolidBlock;
            FToken.InternalLen := Length(Text);
            Inc(FToken.StopCol, FToken.InternalLen);


            if DecodePaired then PairText := Text + TermText;

            while FStream.Position < FStreamSize do
            begin
              FStream.ReadBuffer(P^, 1);
              P[1] := #0;
              if P^ = #13 then
              begin
                Inc(FToken.StopLine);
                DoLineProcessed(FToken.StopLine);
                FToken.StopCol := -2;
              end;

              if DecodePaired and CompareTo(PairText) then
              begin
                FToken.Text := FToken.Text + Text;
                Inc(FToken.InternalLen, Length(Text)*2);
                Inc(FToken.StopCol, Length(Text)*2);
              end
              else
                if CompareTo(TermText) then
                begin
                  FToken.Text := FToken.Text + TermText;
                  Inc(FToken.InternalLen, Length(TermText));
                  Inc(FToken.StopCol, Length(TermText));
                  break;
                end
                else
                begin
                  FToken.Text := FToken.Text + P^;
                  Inc(FToken.StopCol);
                  Inc(FToken.InternalLen);
                end;
            end;
          end;

          if Hidden and (not FEnableHidden) then
          begin
            FToken.Kind := tkError;
            FToken.Text := '';
            FToken.StreamPos := FStream.Position;
            Inc(FToken.StopCol, Length(Text));
            FToken.Col := FToken.StopCol;
            FToken.Line := FToken.StopLine;
            FToken.InternalLen := 0;
            continue;
          end
          else
          begin
            if TermText = '' then
            begin
              FToken.Kind        := tkSymbol;
              FToken.Text        := Text;
              FToken.InternalLen := Length(Text);
              Inc(FToken.StopCol, FToken.InternalLen);
            end;
            break;
          end
        end;
    FToken.Text := FToken.Text + P[0];
    FToken.Kind := tkToken;
    Inc(FToken.InternalLen);
    Inc(FToken.StopCol);
  end;
end;

procedure TStreamTokenizer.GetPrevToken;
begin
  {not implemented}
end;

procedure TStreamTokenizer.Dump(const FileName: String);
var
  i: Integer;
  F: TextFile;
begin
  AssignFile(F, FileName);
  Rewrite(F);
  for i := 0 to FStandAloneTokens.Count-1 do
    with TStandAloneToken(FStandAloneTokens[i]) do
      WriteLn(F, Format('%d->  %s        %s', [i, Text, TermText]));
  CloseFile(F);
end;

end.
